/*
*******************************************************************
*                                                                 *
*                         DiskSims 1.2 Beta                       *
*                    A Disk Schduling Simulator                   *
*       Released under the GPL (GNU General Public Licence)       *
*                                                                 *
*               Copyright (c) 2004-2010 Ammar Muqaddas            *
*                        ammar@solostuff.net                      *
*                        ammar.mu@gmail.com                       *
*                         www.solostuff.net                       *
*                                                                 *
*******************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <memory.h>
#include <math.h>
#include <windows.h>
#include "mt19937int.h"
#define INFINITY 2147483646		//2^31-2
#define USE_SLOOK
#define MAX_MAN_CONF 100

// Enable/Disable rotational latency
#define RLATENCY

// Enable/Disable storing waiting times of individual requests. Allows for extra results to
// generated like variance, but consume extra ammounts of memory.
#define STORE_WAITS
#undef RWAIT
#undef NORWAIT

#define ADD_BPTTIME(bltime,at_time) ((bltime>=at_time) ? bltime : bltime + BPTTIME)

#ifdef RLATENCY
	#define NORWAIT(ALGO,qname) 

	#ifdef STORE_WAITS
		#define RWAIT(ALGO,qname) if (ALGO##_insrv==BLOCKTIME) { w##qname[wi##qname] = cu_time - qname[ALGO##_nexti].atime; ALGO##_wait += w##qname[wi##qname]; if (w##qname[wi##qname]>ALGO##_maxwait) ALGO##_maxwait = w##qname[wi##qname]; else ; wi##qname++; }
	#else
		#define RWAIT(ALGO,qname) if (ALGO##_insrv==BLOCKTIME) { unsigned long tmax = cu_time - qname[ALGO##_nexti].atime; ALGO##_wait += tmax; if (tmax>ALGO##_maxwait) ALGO##_maxwait = tmax; }
	#endif

	// Service time equals block time + rotation time.
	#define SRVTIME(ALGO,qname) BLOCKTIME + ADD_BPTTIME(qname[ALGO##_nexti].bltime,cu_bltime) - cu_bltime
	
	#define MSG_NO_RLATENCY(opt)

	// Enable/Disable SATF
	#define USE_SATF

	// Enable/Disable Advancing current time and rotation when skiping Idle time.
	#undef IDLE_ROT
#else
	#define RWAIT(ALGO,qname) 

	#ifdef STORE_WAITS
		#define NORWAIT(ALGO,qname) { w##qname[wi##qname] = cu_time - qname[ALGO##_nexti].atime; ALGO##_wait += w##qname[wi##qname]; if (w##qname[wi##qname]>ALGO##_maxwait) ALGO##_maxwait = w##qname[wi##qname]; else ; wi##qname++; }
	#else
		#define NORWAIT(ALGO,qname) { a1 = cu_time - qname[ALGO##_nexti].atime; ALGO##_wait += a1; if (a1>ALGO##_maxwait) ALGO##_maxwait = a1; }
	#endif
	
	// Service time equals block time
	#define SRVTIME(ALGO,qname) BLOCKTIME 

	#define MSG_NO_RLATENCY(opt) printf("You should #define RLATENCY to use " #opt " option.\n"); 

	// Always Disable SATF. Don't change this.
	#undef USE_SATF

	// Disable Advancing current time. Don't change this.
	#undef IDLE_ROT
#endif

// Disable/Enable using head settle time.
#define USE_SETTLE

#undef ON_SATF
#ifdef USE_SATF
	#define ON_SATF(exp) exp
#else
	#define ON_SATF(exp)
#endif

#undef ON_SLOOK
#ifdef USE_SLOOK
	#define ON_SLOOK(exp) exp
#else
	#define ON_SLOOK(exp)
#endif


#define GEN_NUMREQ() num_req = (int) (((double) rand()/RAND_MAX) * (max_req_pert+1)); \
					 if (num_req>max_req_pert) num_req=max_req_pert


typedef struct _qreq
	{	int cyl;					// Cylinder number
		unsigned long atime;		// Time the request arrived
		#ifdef RLATENCY
		int bltime;					// (block number)*BLOCKTIME
		#endif
	} qreq;

struct qnode
	{	int reqi;
		int g,h,f,minsi,maxsi;
		int *pp;
		struct qnode *link;
	};

struct qsort_elm
	{	int index,req;
	};

const double GENRAND_MAX = 4294967295;   // 2^32-1
const long WM_PROGRESS = WM_APP + 1, WM_STOP = WM_APP + 2, WM_THREADID = WM_APP + 3;
int MAXCYLINDER = 1000, STARTCYL = 0, BLOCKTIME = 2, CYLSEEK = 1, CYLSEEK_1 = 0 , BPT = -1, BPT_1 = -1, heads = 4;


// Simulation Time
unsigned long cu_time;

#ifdef USE_SETTLE
	int SETTLE=0, SETTLE_CYLSEEK, SETTLE_CYLSEEK_1;		// Head settle time
#endif
#ifdef RLATENCY
	int cu_bltime, BPTTIME;
#endif
int total_requests=-1;
long caller_hwnd = -1, chunk = -1, chunk_count = 0;
unsigned long peek_count = 0;

int (*get_SSTF_nexti)(qreq *q, int SSTF_last, int head);
int (*get_SCAN_nexti)(qreq *q, int SCAN_last, int head, int direc);


int get_SSTF_nexti_nor(qreq *q, int SSTF_last, int head)
{	int n, neari = total_requests; 
	
	for (n=0; n<=SSTF_last; n++)
		if (abs(head-q[n].cyl)<abs(head-q[neari].cyl)) neari=n; 

	if (neari==total_requests) neari=-1;
	return neari;
}

int get_SCAN_nexti_nor(qreq *q, int SCAN_last, int head, int direc)
{	int n,a, neari = total_requests;
	
	for (n=0; n<=SCAN_last; n++)
	{	a=(q[n].cyl-head);
		if (a*direc>=0 && abs(a)<abs(q[neari].cyl-head)) neari=n;
	}
	if (neari==total_requests) neari=-1;
	return neari;
}

__inline void fix_nexti(qreq *q, int last,int head,int direc,int *pnexti)
{	// last is the index of the last request, which is the new request

	int a;
	a=(q[last].cyl-head);
	if ((*pnexti)>-1 && a*direc>=0 && abs(a)<abs(q[*pnexti].cyl-head)) *pnexti=last;
}


#ifdef RLATENCY
int (*get_cyl_nexti)(qreq *q, int last, int neari);

int get_cyl_nexti_rlat(qreq *q, int last, int neari)
{	// A neari is given to the function. Then it tries to get a nearer request to the head 
	// than neari according rotational latency. It doesn't find requests on other cylinders 
	// Note: To work correctly. the head has to be equal to q[neari].cyl. i.e the head is standing
	// on the cylinder of request neari

	int n, cyl; 

	cyl=q[neari].cyl;
	for (n=0; n<=last; n++)
		if (q[n].cyl==cyl && ADD_BPTTIME(q[n].bltime,cu_bltime) < ADD_BPTTIME(q[neari].bltime,cu_bltime)) neari=n;	//**
	
	return neari;
}

int get_cyl_nexti_same(qreq *q, int last, int neari)
{	return neari;
}

int get_SSTF_nexti_rlat(qreq *q, int SSTF_last, int head)
{	
	int n, neari = total_requests, a1; 
	for (n=0; n<=SSTF_last; n++)
	{	a1 = abs(head-q[n].cyl) - abs(head-q[neari].cyl);
		if (a1<0) neari=n; 
		else if (a1==0) 
		{	//int a2;
			a1 = (cu_bltime + abs(head-q[n].cyl) * CYLSEEK) % BPTTIME;
			if (ADD_BPTTIME(q[n].bltime,a1) < ADD_BPTTIME(q[neari].bltime,a1)) neari=n;	//**
		}
	}

	if (neari==total_requests) neari=-1;
	return neari;
}

int get_SCAN_nexti_rlat(qreq *q, int SCAN_last, int head, int direc)
{	int n,a, neari = total_requests, a1;
	for (n=0; n<=SCAN_last; n++)
	{	a=(q[n].cyl-head);
		if (a*direc>=0)
		{	a1 = abs(a) - abs(q[neari].cyl-head);
			if (a1<0) neari=n;
			else if (a1==0)
			{	//int a2;
				a1 = (cu_bltime + abs(a) * CYLSEEK) % BPTTIME;
				if (ADD_BPTTIME(q[n].bltime,a1) < ADD_BPTTIME(q[neari].bltime,a1)) neari=n;
			}
		}
	}
	if (neari==total_requests) neari=-1;
	return neari;
}

#ifdef USE_SATF
int get_SATF_nexti_rlat(qreq *q, int SATF_last, int head)
{	
	int n, neari, neart=INFINITY, a1; 
	for (n=0; n<=SATF_last; n++)
	{	
		#ifdef USE_SETTLE
			if (head!=q[n].cyl)
				a1 = abs(head-q[n].cyl) * CYLSEEK + SETTLE;
			else
				// No Settle time is needed when on the same cylinder. So a1 = 0 + 0
				a1 = 0;			
		#else
			a1 = abs(head-q[n].cyl) * CYLSEEK;
		#endif
		a1 = a1 + ADD_BPTTIME(q[n].bltime,(cu_bltime + a1) % BPTTIME) - (cu_bltime + a1) % BPTTIME;

		if (a1<neart) { neari=n; neart=a1;}
	}

	if (neart==INFINITY) return -1;
	else return neari;
}
#endif

__inline short fix_nexti_rlat(qreq *q, int last,int head,int direc,int *pnexti, long headmv)
{	// Same as fix_nexti but with rotational latency enabled

	int a;
	a=(q[last].cyl-head);
	if ((*pnexti)>-1 && a*direc>=0)
		if (abs(a)<abs(q[*pnexti].cyl-head)) { *pnexti=last; return 1;}
		else if (abs(a)==abs(q[*pnexti].cyl-head))
		{	int a1;
			a1 = (cu_bltime + abs(a) * CYLSEEK + headmv) % BPTTIME;
			if (ADD_BPTTIME(q[last].bltime,a1) < ADD_BPTTIME(q[*pnexti].bltime,a1)) { *pnexti=last; return 2;}
		}
	return 0;
}
#endif

/*
int get_SLOOK_direc(qreq *q, int head, int mini, int maxi)
{	int mid, mad;
	if (q[maxi].cyl==-1) return 0;
	mid=q[mini].cyl-head;
	mad=q[maxi].cyl-head;
	if (abs(mid)<abs(mad))
	{	if (mid>=0) return 1; else return -1;
	} else if (mad>0) return 1; else return -1;
} */

/* int *remove_sim(int *q, int last)
{	int *rq = malloc(sizeof(int)*(last+1));
	int n,n1,rlast=-1;
	short flag1=0;
	for (n=0; n<=last; n++)
	{	flag1=0;
		for (n1=0; n1<=rlast; n1++) flag1=(flag1 || q[n]==rq[n1]);
		if (!flag1) { rlast++; rq[rlast]=q[n] }
	}
} */

int compare(const struct qsort_elm *elem1,const struct qsort_elm *elem2)		//** Definistion is void
{	return (elem1->req)-(elem2->req);
}

void print_listf(struct qnode *plist)
{	while (plist!=NULL)
	{	printf(" %d",plist->f);
		plist=plist->link;
	}
}

void list_finsort(struct qnode **pplist, struct qnode *pitem)
{	if ((*pplist)!=NULL)
		while ((pitem->f)>(*pplist)->f)
		{	pplist=&((*pplist)->link);
			if ((*pplist)==NULL) break;
		}
	pitem->link=*pplist;
	(*pplist)=pitem;
}


__inline void update_progress_bar()
{	if (chunk != -1)
	{	chunk_count++;
		if (chunk_count>=chunk) { PostMessage((HWND) caller_hwnd,WM_PROGRESS,chunk_count,0); chunk_count=0; }
	}
}

/*void CheckArrayEnough(int ar, 
{	// Extends an array if there are not enough elements
	
} */

void main(int argc, char *argv[])
{ 
	int max_req_pert=-1; 
	int rcount = 0, head=-1;
	int n,a,a1,a2,a3,last = -1, qmlast=0,qmi,qmi2;
	char c1, *wcolumn = NULL, *scolumn = NULL, *xwcolumn = NULL;
	long FCFS_head, FCFS_nexti = 0, FCFS_headmv=0;
	long SSTF_head, SSTF_nexti = -1, SSTF_cleft = 0, SSTF_last=last, SSTF_headmv=0;
	long SCAN_head, SCAN_nexti = -1, SCAN_cleft = 0, SCAN_last=last, SCAN_direc=1, SCAN_headmv=0;
	long CSCAN_head, CSCAN_nexti = -1, CSCAN_cleft = 0, CSCAN_last=last, CSCAN_direc=1, CSCAN_headmv=0;
	long LOOK_head, LOOK_nexti = -1, LOOK_cleft = 0, LOOK_last=last, LOOK_direc=1, LOOK_headmv=0;
	long CLOOK_head, CLOOK_nexti = -1, CLOOK_cleft = 0, CLOOK_last=last, CLOOK_direc=1, CLOOK_headmv=0;
	long SLOOK_head, SLOOK_nexti = -1, SLOOK_cleft = 0, SLOOK_last=last, SLOOK_direc=1, SLOOK_headmv=0, SLOOK_mini=0, SLOOK_maxi=0, SLOOK_waist=0, SLOOK_wtfact=-1, SLOOK_wtconst=0; 
	long BOPATH_head, BOPATH_nexti = -1, BOPATH_cleft = 0, BOPATH_last=last, BOPATH_direc=1, BOPATH_headmv=0, BOPATH_optnexti=0, BOPATH_optlast=-1, BOPATH_waist=0, BOPATH_wthresh=-1, BOPATH_maxbnd=0, BOPATH_mini=0, BOPATH_maxi=0; //, BOPATH_mini, BOPATH_maxi,  BOPATH_wtfact; 

	__int64 FCFS_seek = 0, SSTF_seek = 0, SCAN_seek = 0, CLOOK_seek = 0, CSCAN_seek = 0, LOOK_seek = 0, SLOOK_seek = 0, BOPATH_seek = 0;
	__int64 FCFS_wait = 0, SSTF_wait = 0, SCAN_wait = 0, CLOOK_wait = 0, CSCAN_wait = 0, LOOK_wait = 0, SLOOK_wait = 0, BOPATH_wait = 0;
	unsigned long FCFS_maxwait = 0, SSTF_maxwait = 0, SCAN_maxwait = 0, CLOOK_maxwait = 0, CSCAN_maxwait = 0, LOOK_maxwait = 0, SLOOK_maxwait = 0, BOPATH_maxwait = 0;
	short FCFS_insrv = 0, SSTF_insrv = 0, SCAN_insrv = 0, CSCAN_insrv = 0, LOOK_insrv = 0, CLOOK_insrv = 0, SLOOK_insrv = 0, SLOOK_mnmxfl = 0, SLOOK_resmeth=-1, SLOOK_flag1=0 ,BOPATH_insrv = 0, BOPATH_wresetfl=-2, BOPATH_flag1=0, BOPATH_swap=-2, manual = 0, dislabel = 0, skip_idle=0, rlat_method=-1, all_cleft;
	#ifdef USE_SATF
		long SATF_head, SATF_nexti = -1, SATF_cleft = 0, SATF_last=last, SATF_headmv=0;
		__int64 SATF_wait = 0, SATF_seek = 0;
		unsigned long SATF_maxwait = 0;
		short SATF_insrv = 0;
	#endif
	#ifdef RLATENCY
		double geof=-1;
	#endif
	double req_prob=-1, arrival_r=-1, time_unit=-1, poisson_e, poisson_e_1;   // poisson_e = (e^-a);  poisson_e_1 = 1 - poisson_e;
	FILE *wstream = NULL, *sstream = NULL, *xwstream = NULL, *istream=stdin;
	MSG msg1;

	qreq *q;
	qreq *q2;
	qreq *q3;
	qreq *q4;
	qreq *q5;
	qreq *q6;
	qreq *q7 = NULL;
	qreq *q8 = NULL;
	int *qopti = NULL;
	int *qoptinv = NULL;
	qreq *q9 = NULL;
	
	// Waiting Times Arrays
	#ifdef STORE_WAITS
		unsigned long *wq, *wq2, *wq3, *wq4, *wq5, *wq6, *wq7, *wq8;
		int wiq, wiq2, wiq3, wiq4, wiq5, wiq6, wiq7, wiq8;
		#ifdef USE_SATF
			unsigned long *wq9; int wiq9;
		#endif
	#endif


	//int *qtemp;

	// Queus for manual request entry
	int *qm=NULL;
	int *qb=NULL;
	
	srand((unsigned)time(NULL));
	sgenrand((unsigned)time(NULL));
	//srand(600); sgenrand(800);

	for(n=0; n<argc; n++)
		switch (toupper(argv[n][1]))
		{	case 'C':
				if (toupper(argv[n][2])=='W') wcolumn = &argv[n][4];
				else if (toupper(argv[n][2])=='S') scolumn = &argv[n][4];
				else if (toupper(argv[n][2])=='X') xwcolumn = &argv[n][4];
				else chunk=atoi(&(argv[n][3]));
				break;
			case 'W':
				if (toupper(argv[n][2])=='C') { SLOOK_wtconst = atoi(&(argv[n][4])); SLOOK_waist=SLOOK_wtconst; }
				else wstream = fopen(&(argv[n][3]),"a+");
				break;
			case 'S':
				if (toupper(argv[n][2])=='T') 
				#ifdef USE_SETTLE
					SETTLE = atoi(&(argv[n][4]));
				#else 
					printf("\nUSE_SETTLE should be #defined in the code to use the -st option.\n");
				#endif
				else
					sstream = fopen(&(argv[n][3]),"a+");
				break;
			case 'D':
				dislabel = -1;
				break;
			case 'B':
				if ((toupper(argv[n][2])=='P') && (toupper(argv[n][3])=='T'))
				{	//MSG_NO_RLATENCY(-bpt);
					BPT = atoi(&(argv[n][5])); BPT_1 = BPT-1;
				}
				else BLOCKTIME=atoi(&(argv[n][3])); 
				break;
			case 'E':
				CYLSEEK=atoi(&(argv[n][3])); CYLSEEK_1=CYLSEEK-1;
				break;
			case 'M':
				if (toupper(argv[n][2])=='T') 
				{	MSG_NO_RLATENCY(-mt);
					rlat_method = (short) atoi(&(argv[n][4]));
				}
				else MAXCYLINDER=atoi(&(argv[n][3]))-1;
				break;
			case 'R':
				if (toupper(argv[n][2])=='W') SLOOK_resmeth=atoi(&(argv[n][4]));
				else if (toupper(argv[n][2])=='B') BOPATH_wresetfl=-(toupper(argv[n][4])=='Y');
				else if (toupper(argv[n][2])=='S') BOPATH_swap=-(toupper(argv[n][4])=='Y');
				else { srand(atoi(&(argv[n][3]))); sgenrand(atoi(&(argv[n][3]))); }
				break;
			case 'T':
				if (toupper(argv[n][2])=='U') time_unit=atof(&(argv[n][4]));
				else { total_requests=atoi(&(argv[n][3])); manual=(total_requests==0); }
				break;
			case 'N':
				max_req_pert=atoi(&(argv[n][3]));
				break;
			case 'P':
				req_prob=atof(&(argv[n][3]));
				//else if (argv[n][2]=='2') req_prob2=atof(&(argv[n][4]));
				//else if (argv[n][2]=='3') req_prob3=atof(&(argv[n][4]));
				break;
			case 'I':
				if (toupper(argv[n][2])=='N') istream = fopen(&(argv[n][4]),"r");
				else if (toupper(argv[n][2])=='D') skip_idle=-1;
				else head=atoi(&(argv[n][3]));
				break;
			case 'F':
				if ((toupper(argv[n][2])=='C') && (toupper(argv[n][3])=='F') && (toupper(argv[n][4])=='S'))
					FCFS_nexti=-2;
				else SLOOK_wtfact=atoi(&(argv[n][3]));
				break;
			case 'O':
				BOPATH_wthresh=atoi(&(argv[n][3]));
				break;
			case 'X':
				if (toupper(argv[n][2])=='W') xwstream = fopen(&(argv[n][4]),"a+");
				else BOPATH_maxbnd=atoi(&(argv[n][3]));
				break;
			case 'A':
				arrival_r = atof(&(argv[n][3]));
				break;
			case '?':
				; 
			case 'H':
				if (toupper(argv[n][2])=='W') 
				{	caller_hwnd = atoi(&(argv[n][4]));
					PostMessage((HWND) caller_hwnd,WM_THREADID,GetCurrentThreadId(),0);
				}
				else if (toupper(argv[n][2])=='D') 
				{	heads = atoi(&(argv[n][4]));
				}
				else {
				printf("\nDiskSims v1.2 Beta. A Disk Scheduling Simulator\nCopyright (c) 2005-2010 Ammar Muqaddas\n\n");
				printf("Usage:  schedule [-w:waittime_file -cw:column] [-s:seek_file -cs:column] [-xw:maxwait_file -cx:column] [-d] [-in:infile] [-b:blocktime] [-e:seektime] [-st:settle] [-m:cylinders] [-bpt:bpt] [-r:seed] [-t:total] [-n:MNRPTU] [-a:rate -tu:time] [-p:probability] [-i:head] [-f:factor] [-rw:y/n] [-o:threshold] [-x:maxdepth] [-mt:(1/2)] [-hw:hwnd [-c:chunk]] [-id]\n\n");
				printf("   -w:waittime_file     Specifies an output csv file for waiting time results.\n");
				printf("   -s:seek_file         Specifies an output csv file for seek time results.\n");
				printf("   -xw:maxwait_file     Specifies an output csv file for maximum waiting time\n");
				printf("                        results.\n");
				printf("   -cw:column           A value that will be writen to the first column of\n                        the next row in <waittime_file>.\n");
				printf("   -cs:column           A value that will be writen to the first column of\n                        the next row in <seek_file>.\n");
				printf("   -cx:column           A value that will be writen to the first column of\n                        the next row in <maxwait_file>.\n");
				printf("   -d                   Disable adding column labels in new or empty files.\n");
				printf("   -in:infile           Specifies an input file. The program will take it's\n                        input from this file instead of asking the user.\n");
				printf("   -b:blocktime         Single block transfer time (in time units)\n");
				printf("   -e:seektime          Head's single cylinder movement time (in time units)\n");
				printf("   -st:settle           Head settle time.\n");
				printf("   -m:cylinders         Total number of cylinders in hard disk.\n");
				printf("   -bpt:bpt             Number of blocks per track. RLATENCY should be defined.\n");         
				printf("   -hd:heads            Number of heads.\n"); 
				printf("   -r:seed              Randomization seed. Ensures that same requests are\n                        generated on each run. If omitted, The timer is used\n                        which results in truly random request generation.\n");
				printf("   -t:total             Total number of requests to be generated.\n");
				printf("   -n:MNRPTU            Maximum number of requests per time unit.\n");
				printf("   -a:rate              Requests arrival rate per second. Used for Poisson                              Distribution. If used -p and -n are ignored.\n");
				printf("   -tu:time             Seconds per time unit. Used for Poisson Distribution.                           If used -p and -n are ignored.\n");
				printf("   -p:probability       Prbability of whether to generate requests or not\n                        in each time unit.\n");
				printf("   -i:head              Head's initial position. (cylinder number).\n");
				printf("   -f:factor            S-LOOK waisted time factor.\n");
				printf("   -wc:constant         S-LOOK waisted time constant\n");				
				printf("   -rw:(1/2/3)          Should be 1, 2 or 3. SLOOK waisted time resting method\n                          1. Extremes: Reset on every min/max\n                          2. Ping Pong: Reset on min if the previous reset is\n                             max and reset on max if previous reset is min\n                          3. Empty Queue: Reset when request queue is empty\n");
				printf("   -o:threshold         BO-PATH waisted time threshold.\n");
				printf("   -x:maxdepth	        BO-PATH maximum A* depth (Min=1).\n");
				//printf("   -rb:(y/n)            Should be 'yes' or 'no'. Indicates that BO-PATH\n                        waisted time should be reset on every request in                                path.\n");
				//printf("   -rs:(y/n)            Should be 'yes' or 'no'. Use BO-PATH min/max\n                        swaping trick.\n");
				printf("   -mt:(1/2)            Should be 1 or 2. Rotational Latency method for\n");
				printf("                        dealing with requests on the same cylinder.\n");
				//printf("                        the closest request immediately\n");
				printf("   -hw:hwnd             (For system use). Window handle of calling process.\n");
				printf("   -c:chunk             (For system use). Number of messages to collect\n                        before sending a chunked progress message.\n");
				printf("   -id                  Skip idle simulation time to improve performance.\n                        Idle time is when all heads of all algorithms \n                        are not doing any thing. Usefull for very low request\n                        probabilities.");
				exit(0); 
				}
		}
	if (arrival_r<=0 || time_unit<=0) { arrival_r=-1; time_unit=-1; }
	else max_req_pert=0;

	if (BPT==-1) { BPT=600; BPT_1=599; }
	#ifdef RLATENCY
		// Setting some default values if they are not specified.
		if (rlat_method==-1) rlat_method=1;

	#else
		// Checking for options that are NOT enabled by #defining compiler directives
		//if (caller_hwnd==-1 && (BPT!=-1 || rlat_method!=-1))
		if (caller_hwnd==-1 && (rlat_method!=-1))
			exit(0);
	#endif

		
	// Writing Column Labels
	if (!dislabel)
	{	if (wstream!=NULL)
		{	fseek(wstream,0,SEEK_END);
			if (ftell(wstream)==0) fprintf(wstream,",FCFS,SSTF,SCAN,C-SCAN,LOOK,C-LOOK,S-LOOK,BO-PATH,SATF\n");
		}
		if (sstream!=NULL)
		{	fseek(sstream,0,SEEK_END);
			if (ftell(sstream)==0) fprintf(sstream,",FCFS,SSTF,SCAN,C-SCAN,LOOK,C-LOOK,S-LOOK,BO-PATH,SATF\n");
		}
		if (xwstream!=NULL)
		{	fseek(xwstream,0,SEEK_END);
			if (ftell(xwstream)==0) fprintf(xwstream,",FCFS,SSTF,SCAN,C-SCAN,LOOK,C-LOOK,S-LOOK,BO-PATH,SATF\n");
		}
	}

	//printf("%s\n",argv[n]);


	// Entering values
	if (total_requests==-1)
	{	printf("\nEnter total number of requests (0=Manual Entry): ");
		fscanf(istream,"%d", &total_requests); manual=(total_requests==0);
	}
	if (manual)
	{	short useLBA=0;
		char pin[MAX_MAN_CONF];
		int BPTXheads = BPT * heads;
	
		a=5000; a1=0; total_requests=0; qmi=0;
		qm=realloc(qm,a*sizeof(int));
		#ifdef RLATENCY
			qb=realloc(qb,a*sizeof(int));
		#endif

		// Read manual configuration line
		printf("Enter configuration line or press ENTER to skip: "); 
		a2=getc(istream); if (a2!='\n') ungetc(a2,istream);
		fgets(pin,MAX_MAN_CONF,istream);

		// Load the values
		sscanf(pin,"%hd",&useLBA);
		if (useLBA) printf("\nManual requests using LBA...\n");


		// Begin manual entry loop
		printf("Enter requests for t=%d (-1=stop): ",a1); 
		do
		{	do
			{	a2=getc(istream);
				if (a2==',' || a2=='t') 
				{		
					qmlast++; 
					// Check if array size is enough and increase if not.
					if ((qmlast+1)>a) 
					{	a=a+5000; qm=realloc(qm,a*sizeof(int)); 
						#ifdef RLATENCY
							qb=realloc(qb,a*sizeof(int)); 
						#endif
					}
					
					// If it's a time delay. (Time delays are expressed in negative, -1 is reserved. So time delay = -real_delay -1
					if (a2=='t')
					{	fscanf(istream,"%d",&qm[qmlast]);
						a1 = a1 + qm[qmlast];
						qm[qmlast] = -qm[qmlast] - 1;
					}
					else 
					{	a1++; 
						qm[qmlast]=-2; //qb[qmlast]=-2;
					}

				}
				else if (a2=='\n' && istream==stdin)  // No need to print this is istream!=stdin
					printf("Enter requests for t=%d (-1=stop): ",a1); 
			} while (a2==' ' || a2=='\n' || a2==',' || a2=='t');
			

			qmlast++; total_requests++; 
			if (a2==EOF) break;
			ungetc(a2,istream);

			// Check if array size is enough and increase if not.
			if ((qmlast+1)>a) 
			{	a=a+5000; qm=realloc(qm,a*sizeof(int)); 
				#ifdef RLATENCY
					qb=realloc(qb,a*sizeof(int)); 
				#endif
			}

			// If LBA is used, convert it to CHS
			fscanf(istream,"%d",&a3);
			if (useLBA)
			{
				// First, just check that a2 is not ':'. If so stop the program, 
				// obiuosly there is no meaning of LBA while using sector values
				if (a2==':') 
				{	printf("\n\nInput file uses LBA, but some requests are using block numbers!!");
					exit(0);
				}

				qm[qmlast] = a3 / BPTXheads;

				#ifdef RLATENCY
					qb[qmlast] = a3 % BPT;
				#endif
			} else 
			{
			// If LBA is NOT used, get the cylinder number and block number
				qm[qmlast] = a3;
				if (a3!=-1)
				{
					// Bypass a delimiter ':' if it exists reading the block number.
					a2=getc(istream); 
					if (a2!=':') 
					{	ungetc(a2,istream);
						#ifdef RLATENCY
							// Since there is no block number, assume it's zero 
							qb[qmlast]=0;
						#endif
					}
					else 
					{	fscanf(istream,"%d",&a2);
						#ifdef RLATENCY
							// Get the block nmuber
							qb[qmlast]=a2;
						#endif
					}
				}
			}
			//printf(" %d",a3);
		} while (a3!=-1);

		// Mark begining with -1 instead of -2
		qm[0] = -1;

		// Make sure that end is marked with -1
		qm[qmlast] = -1;
		
		//#ifdef RLATENCY
		//	for (n=0; n<=qmlast; n++)
		//		printf(" %d:%d",qm[n],qb[n]);
		//#else
		//	for (n=0; n<=qmlast; n++)
		//		printf(" %d",qm[n]);
		//#endif
		qmlast--; total_requests--;
		printf("\nTotal number of requests=%d\n\n",total_requests);
	} else
	{	if (max_req_pert==-1)
		{	printf("\nEnter Maximum number of requests per time unit (0=Generate All in t=0): ");
			fscanf(istream,"%d", &max_req_pert);
		}
		if (max_req_pert)
		{	if (req_prob==-1)
			{	printf("\nEnter request probability %%: ");		// Minimum non-zero probability is 1/(GENRAND_MAX+1)
				if (req_prob!=-1) printf("%g ", req_prob); 
				if (req_prob==-1) fscanf(istream," %lg", &req_prob); 
			}
			//if ((req_prob<0.00306 || req_prob2<0.00306 || req_prob3<0.00306) && req_prob!=0 && req_prob2!=0 && req_prob3!=0) 
			//	{ req_prob=0; printf(">> One of the probabilities is too low, Assumming 0%%\n"); 
			//	}
			req_prob = req_prob/100; printf("\n>> %.10f\n", req_prob);
			#ifdef RLATENCY
				geof = 1/log(1-req_prob);
			#endif
		} else req_prob=0;
	}  // if Manual

	if (head==-1)
	{	printf("Enter head's initial position: ");
		fscanf(istream,"%d", &a);
	} else a=head;
	head=a; FCFS_head=a; SSTF_head=a; SCAN_head=a; CSCAN_head=a; LOOK_head=a; CLOOK_head=a; SLOOK_head=a; BOPATH_head=a; ON_SATF(SATF_head=a);
	if (SLOOK_wtfact==-1)
	{	printf("\nEnter S-LOOK waisted time factor: ");
		fscanf(istream,"%d", &SLOOK_wtfact);
	}
	if (SLOOK_resmeth==-1)
	{	printf("\nChoose SLOOK waisted time resting method:\n");
		printf("1. Reset on every min/max\n");
		printf("2. Reset on min if the previous is max and reset on max if previous is min\n");
		printf("3. Reset when request queue is empty (Default)\n");
		//fscanf(istream," \n\t%c", &c1); SLOOK_resmeth=-(c1=='y');
		fscanf(istream,"%d",&SLOOK_resmeth);
	}
	if (BOPATH_wthresh==-1)
	{	printf("\nEnter BO-PATH waisted time threshold: ");
		fscanf(istream,"%d",&BOPATH_wthresh);
	}
	while (BOPATH_maxbnd<1)
	{	printf("\nEnter BO-PATH maximum A* depth (Min=1): ");
		fscanf(istream,"%d",&BOPATH_maxbnd);
		if (BOPATH_maxbnd<1) printf(">> Too low, Minimum = 1");
	}
	BOPATH_maxbnd--;
	if (BOPATH_wresetfl==-2)
	{	printf("\nReset BO-PATH waisted time on every request in path ? (y/n) ");
		fscanf(istream," \n\t%c", &c1); BOPATH_wresetfl=-(c1=='y');
	}
	if (BOPATH_swap==-2)
	{	printf("\nUse BO-PATH min/max swaping trick ? (y/n) ");
		fscanf(istream," \n\t%c", &c1); BOPATH_swap=-(c1=='y');
	}


	// Preparing queues
	q=malloc(sizeof(qreq)*(total_requests+1));			// For FCFS
	q2=malloc(sizeof(qreq)*(total_requests+1));			// For SSTF
	q2[total_requests].cyl= 5*MAXCYLINDER;
	q3=malloc(sizeof(qreq)*(total_requests+1));			// For SCAN
	q3[total_requests].cyl= 5*MAXCYLINDER;
	q4=malloc(sizeof(qreq)*(total_requests+1));			// For C-SCAN
	q4[total_requests].cyl= 5*MAXCYLINDER;
	q5=malloc(sizeof(qreq)*(total_requests+1));			// For LOOK
	q5[total_requests].cyl= 5*MAXCYLINDER;
	q6=malloc(sizeof(qreq)*(total_requests+1));			// For C-LOOK
	q6[total_requests].cyl= 5*MAXCYLINDER;
	#ifdef USE_SLOOK
		q7=malloc(sizeof(qreq)*(total_requests+2));			// For S-LOOK
		q7[total_requests].cyl= 5*MAXCYLINDER; q7[total_requests+1].cyl= -1;
		SLOOK_mini=total_requests; SLOOK_maxi=total_requests+1;
	#endif
	#ifdef USE_SATF
		q9=malloc(sizeof(qreq)*(total_requests+1));		// For SATF
		q9[total_requests].cyl= 5*MAXCYLINDER;
	#endif

	#ifdef STORE_WAITS
		wq=malloc(sizeof(*wq)*total_requests);			// For FCFS
		wq2=malloc(sizeof(*wq2)*total_requests);		// For SSTF
		wq3=malloc(sizeof(*wq3)*total_requests);		// For SCAN
		wq4=malloc(sizeof(*wq4)*total_requests);		// For C-SCAN
		wq5=malloc(sizeof(*wq5)*total_requests);		// For LOOK
		wq6=malloc(sizeof(*wq6)*total_requests);		// For C-LOOK
		wq7=malloc(sizeof(*wq7)*total_requests);		// For S-LOOK
		wq8=malloc(sizeof(*wq8)*total_requests);		// For BO-PATH
		
		// Last serviced request indexes.
		wiq=0; wiq2=0; wiq3=0; wiq4=0; wiq5=0; wiq6=0; wiq7=0; wiq8=0;
		
		#ifdef USE_SATF
			wq9=malloc(sizeof(*wq9)*total_requests);	// For SATF
			wiq9=0;
		#endif
	#endif


	// Disable FCFS option
	if (FCFS_nexti==-2) FCFS_nexti=total_requests+10;
	
	#ifdef RLATENCY
		BPTTIME = BPT*BLOCKTIME; cu_bltime=0;

		// Choosing Rotational latency optimization method 
		if (rlat_method==1) 
		{	get_SSTF_nexti = &get_SSTF_nexti_rlat;
			get_SCAN_nexti = &get_SCAN_nexti_rlat;
			get_cyl_nexti = &get_cyl_nexti_same;
		} else  // =2
		{	get_SSTF_nexti = &get_SSTF_nexti_nor;
			get_SCAN_nexti = &get_SCAN_nexti_nor;
			get_cyl_nexti = &get_cyl_nexti_rlat;
		} 
	#else
		get_SSTF_nexti = &get_SSTF_nexti_nor;
		get_SCAN_nexti = &get_SCAN_nexti_nor;
	#endif



	// Start Simulation
	#ifdef USE_SETTLE
		SETTLE_CYLSEEK = SETTLE+CYLSEEK; SETTLE_CYLSEEK_1 = SETTLE+CYLSEEK_1;
	#endif
	peek_count = GetTickCount(); cu_time=0;
	poisson_e = exp(-arrival_r * time_unit); poisson_e_1=1-poisson_e;
	printf("\nSimulation Started....\n");
	all_cleft = SSTF_cleft || FCFS_nexti<=last || SCAN_cleft || CSCAN_cleft || LOOK_cleft || CLOOK_cleft || SLOOK_cleft || BOPATH_cleft ON_SATF(|| SATF_cleft);

	while ((rcount<total_requests) || all_cleft)
	{	int num_req;
		double d1;

		// Checking if caller wants to stop the simulation
		if ((GetTickCount()>peek_count+200) || GetTickCount()<peek_count) if (PeekMessage(&msg1,NULL,WM_STOP,WM_STOP,PM_REMOVE)) goto EXIT_MAIN; else peek_count=GetTickCount();
		//if (PeekMessage(&msg1,NULL,WM_STOP,WM_STOP,PM_REMOVE)) goto EXIT_MAIN;

		// Choosing number of req. to be generated
		if (!manual)
		{	
			// Possion Random Variable Generator.
			if (arrival_r!=-1)
			{	
				// Skip Idle time
				if (skip_idle && !all_cleft && CSCAN_head!=MAXCYLINDER && CSCAN_direc!=-1)
				{	num_req=0; d1 = poisson_e + ((double) genrand()/GENRAND_MAX)*poisson_e_1;
					#ifdef IDLE_ROT
						// Advance current time and rotation accordingly
						cu_bltime = (cu_bltime - (int) (log(1-((double) genrand()/GENRAND_MAX))/(arrival_r*time_unit))) % BPTTIME;
					#endif
				}
				else { num_req=-1; d1=1; }
				do 
				{	num_req++;
					d1 = d1 * ((double) genrand()/GENRAND_MAX); 
				}
				while (d1>=poisson_e);
			}
			
			// Generic Random Variable Generator.
			else if (max_req_pert)
			{	if (genrand()<=req_prob*(GENRAND_MAX+1)-1)
				{	GEN_NUMREQ();
				} 
				else
				// Skip Idle Time
				if (skip_idle && !all_cleft && CSCAN_head!=MAXCYLINDER && CSCAN_direc!=-1)
				{	GEN_NUMREQ();
					#ifdef IDLE_ROT
						// Advance current time and rotation accordingly. Produce a geometric RV to estimate time passed.
						cu_bltime = (cu_bltime + (int) ceil(geof*log(1-(double) genrand()/GENRAND_MAX)-1) + 1) % BPTTIME;
					#endif
				} 
				else num_req = 0;
			} else num_req=total_requests;
		
		} else  //Manual
		{	num_req=0;
			
			// Reduce time left in a time delay (t). This is done by increasing by 1 since it's in negative.
			if (qm[qmi]<-1) qm[qmi]++;
			else 
			{	
				// If not a time delay proceed normal adding requests.
				if (qmi<=qmlast) { qmi++; qmi2=qmi; }
				while (qm[qmi]>=0)
				{	num_req++; qmi++;
				}
				
			}
		}
		
		rcount = rcount + num_req; 
		if (rcount > total_requests)
		{	num_req = num_req + total_requests - rcount;
			rcount = total_requests;
		}

		// Generating requests
		for (n=1; n<=num_req; n++)
		{	//printf("r");
			last++, SSTF_last++; SSTF_cleft++; SCAN_last++; SCAN_cleft++; CSCAN_last++; CSCAN_cleft++; 
			LOOK_last++; LOOK_cleft++; CLOOK_last++; CLOOK_cleft++; ON_SLOOK(SLOOK_last++; SLOOK_cleft++); 
			#ifdef USE_SATF
				SATF_last++; SATF_cleft++;
			#endif
			if (manual)
			{	q[last].cyl=qm[qmi2];
				#ifdef RLATENCY
					// Set Block number (still not maped to block time)
					q[last].bltime = qb[qmi2];
				#endif
				qmi2++;
			} else 
			{	q[last].cyl = (int) (((double) rand()/RAND_MAX) * (MAXCYLINDER+1));
				#ifdef RLATENCY
					// Set Block number (still not maped to block time)
					q[last].bltime = (int) (((double) genrand()/GENRAND_MAX) * BPT);
				#endif
			}
			
			// Fixing out of bounds values. Very unlikely to happen
			if (q[last].cyl>MAXCYLINDER) q[last].cyl=MAXCYLINDER;
			#ifdef RLATENCY
				if (q[last].bltime==BPT) q[last].bltime = BPT_1;
			
				// Map the block number to it's block time.
				q[last].bltime *= BLOCKTIME;
			#endif

			// Setting the new request for all the Qs
			//if ((insort(q2,last-1,q[last].cyl)<=SSTF_nexti) && (SSTF_nexti>-1)) SSTF_nexti++;
			q2[SSTF_last].cyl=q[last].cyl; q3[SCAN_last].cyl=q[last].cyl; q4[CSCAN_last].cyl=q[last].cyl; q5[LOOK_last].cyl=q[last].cyl; q6[CLOOK_last].cyl=q[last].cyl; 

			// Setting request arrival time
			q2[SSTF_last].atime=cu_time; q3[SCAN_last].atime=cu_time; q4[CSCAN_last].atime=cu_time; q5[LOOK_last].atime=cu_time; q6[CLOOK_last].atime=cu_time; 

			#ifdef USE_SATF
				q9[SATF_last].cyl=q[last].cyl; q9[SATF_last].atime=cu_time;
			#endif
			#ifdef USE_SLOOK
				q7[SLOOK_last].cyl=q[last].cyl; q7[SLOOK_last].atime=cu_time; 
			#endif

			#ifdef RLATENCY
				q2[SSTF_last].bltime=q[last].bltime; q3[SCAN_last].bltime=q[last].bltime; q4[CSCAN_last].bltime=q[last].bltime; q5[LOOK_last].bltime=q[last].bltime; q6[CLOOK_last].bltime=q[last].bltime; 
				#ifdef USE_SATF
					q9[SATF_last].bltime=q[last].bltime;
				#endif
				#ifdef USE_SLOOK
					q7[SLOOK_last].bltime=q[last].bltime;
				#endif
			#endif


			#ifdef USE_SLOOK
				// SLOOK, Fixing mini and maxi
				if (q7[SLOOK_last].cyl>q7[SLOOK_maxi].cyl) { SLOOK_maxi=SLOOK_last; SLOOK_mnmxfl=-1; }
				if (q7[SLOOK_last].cyl<q7[SLOOK_mini].cyl) { SLOOK_mini=SLOOK_last; SLOOK_mnmxfl=-1; }
			#endif


			#ifdef RLATENCY
				if (rlat_method==1)
				{	// Fixing SCAN nexti
					if (fix_nexti_rlat(q3,SCAN_last,SCAN_head,SCAN_direc,&SCAN_nexti,SCAN_headmv)==2 && SCAN_insrv) 
						SCAN_insrv = SRVTIME(SCAN,q3);	

					// Fixing C-SCAN nexti
					if (CSCAN_direc==1)
						if (fix_nexti_rlat(q4,CSCAN_last,CSCAN_head,CSCAN_direc,&CSCAN_nexti,CSCAN_headmv)==2 && CSCAN_insrv) 
							CSCAN_insrv = SRVTIME(CSCAN,q4);

					// Fixing LOOK nexti
					if (fix_nexti_rlat(q5,LOOK_last,LOOK_head,LOOK_direc,&LOOK_nexti,LOOK_headmv)==2 && LOOK_insrv) 
						LOOK_insrv = SRVTIME(LOOK,q5);

					// Fixing C-LOOK nexti
					if (CLOOK_direc==1)
						if (fix_nexti_rlat(q6,CLOOK_last,CLOOK_head,1,&CLOOK_nexti,CLOOK_headmv)==2 && CLOOK_insrv)
							CLOOK_insrv = SRVTIME(CLOOK,q6); else ;
					else if (fix_nexti_rlat(q6,CLOOK_last,STARTCYL,1,&CLOOK_nexti,CLOOK_headmv)==2 && CLOOK_insrv)
							CLOOK_insrv = SRVTIME(CLOOK,q6);

					#ifdef USE_SLOOK
						// Fixing SLOOK nexti
						if (fix_nexti_rlat(q7,SLOOK_last,SLOOK_head,SLOOK_direc,&SLOOK_nexti,SLOOK_headmv)==2 && SLOOK_insrv)
							SLOOK_insrv = SRVTIME(SLOOK,q7);
					#endif

				} else 
			#endif
				{	// Fixing SCAN nexti
					fix_nexti(q3,SCAN_last,SCAN_head,SCAN_direc,&SCAN_nexti);
				
					// Fixing C-SCAN nexti
					if (CSCAN_direc==1)
						fix_nexti(q4,CSCAN_last,CSCAN_head,CSCAN_direc,&CSCAN_nexti);

					// Fixing LOOK nexti
					fix_nexti(q5,LOOK_last,LOOK_head,LOOK_direc,&LOOK_nexti);

					// Fixing C-LOOK nexti
					if (CLOOK_direc==1)
						fix_nexti(q6,CLOOK_last,CLOOK_head,1,&CLOOK_nexti);
					else fix_nexti(q6,CLOOK_last,STARTCYL,1,&CLOOK_nexti);

					// Fixing SLOOK nexti
					ON_SLOOK(fix_nexti(q7,SLOOK_last,SLOOK_head,SLOOK_direc,&SLOOK_nexti));

				}

		}

		#ifdef USE_SLOOK
			// SLOOK, Choosing shortest direction
			if (SLOOK_mnmxfl && SLOOK_cleft && !SLOOK_headmv && !SLOOK_insrv && SLOOK_waist!=INFINITY)
			{	SLOOK_mnmxfl=0;
				a=q7[SLOOK_mini].cyl-SLOOK_head; a1=q7[SLOOK_maxi].cyl-SLOOK_head;
				if (abs(a)<abs(a1))
				{	if (a>=0) n=1; else n=-1;
					if (n!=SLOOK_direc && abs(a)+SLOOK_waist<abs(a1)) { SLOOK_direc=n; SLOOK_nexti=-1; /*printf("*");*/ }
				} else
				{	if (a1>0) n=1; else n=-1;
					if (n!=SLOOK_direc && abs(a1)+SLOOK_waist<abs(a)) { SLOOK_direc=n; SLOOK_nexti=-1; /*printf("*");*/ }
				}
			}
		#endif


		// FCFS
again2:	if (FCFS_headmv)
		{	FCFS_headmv--;	
			if (rcount==total_requests)
			{	FCFS_wait = FCFS_wait + (last-FCFS_nexti+1)*FCFS_headmv; 
				FCFS_headmv=0;
			}
		}
		else if (FCFS_insrv)
		{	RWAIT(FCFS,q);
			FCFS_insrv--; if (!FCFS_insrv) FCFS_nexti++; 
		} 
		else
		{	if (FCFS_nexti<=last)
			{	n = q[FCFS_nexti].cyl-FCFS_head;
				if (n!=0)
				{	n=n/abs(n); FCFS_head = FCFS_head + n;
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (q[FCFS_nexti].cyl==FCFS_head)
						{	FCFS_seek+=SETTLE_CYLSEEK; FCFS_headmv=SETTLE_CYLSEEK_1; }
						else
					#endif
						{ FCFS_seek+=CYLSEEK; FCFS_headmv=CYLSEEK_1; }
				
				} else { NORWAIT(FCFS,q); FCFS_insrv = SRVTIME(FCFS,q); update_progress_bar(); goto again2; }
			}
		}
		//FCFS_wait = FCFS_wait+last-FCFS_nexti+1; 

		
		//if ((rcount==total_requests) && SSTF_cleft==3) { printf("\n^^^%d\n",SSTF_cleft); scanf("%d",&n); }
		//if (SSTF_cleft<0) { printf("\n***%d\n",SSTF_cleft); scanf("%d",&n); }


		//SSTF
again1:	if (SSTF_headmv)
			SSTF_headmv--;
		else if (SSTF_insrv)
		{	//if ((SSTF_insrv==BLOCKTIME) && cu_bltime==q2[SSTF_nexti].bltime) printf("*");
			RWAIT(SSTF,q2);
			SSTF_insrv--; if (!SSTF_insrv) { q2[SSTF_nexti]=q2[SSTF_last]; SSTF_last--; SSTF_cleft--; SSTF_nexti = -1; } 
		} 
		else 
		{	if (SSTF_nexti==-1) SSTF_nexti = get_SSTF_nexti(q2,SSTF_last,SSTF_head);
			if (SSTF_nexti>-1)
			{	n = q2[SSTF_nexti].cyl-SSTF_head;
				if (n!=0)
				{	n=n/abs(n); SSTF_head = SSTF_head + n;
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (q2[SSTF_nexti].cyl==SSTF_head)
						{	SSTF_seek+=SETTLE_CYLSEEK; SSTF_headmv=SETTLE_CYLSEEK_1; }
						else
					#endif
						{ SSTF_seek+=CYLSEEK; SSTF_headmv=CYLSEEK_1;
						}
				} else 
				{	
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						SSTF_nexti = get_cyl_nexti(q2,SSTF_last, SSTF_nexti);
					#endif
					NORWAIT(SSTF,q2);
					SSTF_insrv = SRVTIME(SSTF,q2); update_progress_bar(); goto again1;
				}
			}
		}
		//SSTF_wait = SSTF_wait+SSTF_cleft;


		//SCAN
again3:	if (SCAN_headmv)
			SCAN_headmv--;
		else if (SCAN_insrv)
		{	RWAIT(SCAN,q3);
			SCAN_insrv--; if (!SCAN_insrv) { q3[SCAN_nexti]=q3[SCAN_last]; SCAN_last--; SCAN_cleft--; SCAN_nexti = -1; } //else RWAIT(SCAN);
		} 
		else 
		{	if (SCAN_nexti==-1) SCAN_nexti = get_SCAN_nexti(q3,SCAN_last,SCAN_head,SCAN_direc);
			if (SCAN_nexti>-1)
				if (q3[SCAN_nexti].cyl==SCAN_head)
				{ 
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						SCAN_nexti = get_cyl_nexti(q3,SCAN_last, SCAN_nexti);
					#endif
					NORWAIT(SCAN,q3);
					SCAN_insrv = SRVTIME(SCAN,q3); update_progress_bar(); goto again3;
				}
			if (SCAN_cleft) 
			{	if ((SCAN_head==STARTCYL && SCAN_direc==-1) || (SCAN_head==MAXCYLINDER && SCAN_direc==1))
				{	SCAN_direc = -SCAN_direc;
					#ifdef USE_SETTLE
						SCAN_nexti = get_SCAN_nexti(q3,SCAN_last,SCAN_head,SCAN_direc);
					#endif
				}
				SCAN_head = SCAN_head + SCAN_direc; 
				
				#ifdef USE_SETTLE
					// If the head has just arrived at the destination cylinder then add "settle time"
					if (SCAN_nexti>-1 && q3[SCAN_nexti].cyl==SCAN_head)
					{	SCAN_seek+=SETTLE_CYLSEEK; SCAN_headmv=SETTLE_CYLSEEK_1; }
					else
				#endif
					{	SCAN_seek+=CYLSEEK; SCAN_headmv=CYLSEEK_1; }
			}
		}
		//SCAN_wait = SCAN_wait+SCAN_cleft;


		//C-SCAN
again4:	if (CSCAN_headmv)
			CSCAN_headmv--;
		else if (CSCAN_insrv)
		{	RWAIT(CSCAN,q4);
			CSCAN_insrv--; if (!CSCAN_insrv) { q4[CSCAN_nexti]=q4[CSCAN_last]; CSCAN_last--; CSCAN_cleft--; CSCAN_nexti = -1; } //else RWAIT(CSCAN);
		} 
		else 
		{	if (CSCAN_direc==1)
			{	if (CSCAN_nexti==-1) CSCAN_nexti = get_SCAN_nexti(q4,CSCAN_last,CSCAN_head,CSCAN_direc);
				if (CSCAN_nexti>-1)
					if (q4[CSCAN_nexti].cyl==CSCAN_head)
					{ 
						#ifdef RLATENCY
							// If rlat_method=2, this will search for the nearest request in current cylinder.
							CSCAN_nexti = get_cyl_nexti(q4,CSCAN_last, CSCAN_nexti);
						#endif
						NORWAIT(CSCAN,q4);
						CSCAN_insrv = SRVTIME(CSCAN,q4); update_progress_bar(); goto again4;
					}
				if (CSCAN_cleft || CSCAN_head==MAXCYLINDER) 
				{	if (CSCAN_head==MAXCYLINDER) 
					{	CSCAN_direc=-1;
						CSCAN_head+=CSCAN_direc; CSCAN_seek+=CYLSEEK; CSCAN_headmv=CYLSEEK_1;
					} else 
					{	CSCAN_head+=CSCAN_direc;
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (CSCAN_nexti>-1 && q4[CSCAN_nexti].cyl==CSCAN_head)
						{	CSCAN_seek+=SETTLE_CYLSEEK; CSCAN_headmv=SETTLE_CYLSEEK_1; }
						else 
					#endif
						{	CSCAN_seek+=CYLSEEK; CSCAN_headmv=CYLSEEK_1; }
					}
				}
			}
			else 
			{	if (CSCAN_head==STARTCYL) 
				{	CSCAN_direc=1; 
					#ifdef USE_SETTLE
						// Special case for settle time. When there is a request at the first cylinder
						CSCAN_nexti = get_SCAN_nexti(q4,CSCAN_last,CSCAN_head,1);
						if (CSCAN_nexti>-1 && q4[CSCAN_nexti].cyl==CSCAN_head)
						{	CSCAN_seek+=SETTLE; CSCAN_headmv=SETTLE; }
					#endif
					goto again4; 
				}
				CSCAN_head+=CSCAN_direc; CSCAN_seek+=CYLSEEK; CSCAN_headmv=CYLSEEK_1;
			}
		}
		//CSCAN_wait = CSCAN_wait+CSCAN_cleft;


		//LOOK
again5:	if (LOOK_headmv)
			LOOK_headmv--;
		else if (LOOK_insrv)
		{	RWAIT(LOOK,q5);
			LOOK_insrv--; if (!LOOK_insrv) { q5[LOOK_nexti]=q5[LOOK_last]; LOOK_last--; LOOK_cleft--; LOOK_nexti = -1; } //else RWAIT(LOOK);
		} 
		else
		{	if (LOOK_nexti==-1) LOOK_nexti = get_SCAN_nexti(q5,LOOK_last,LOOK_head,LOOK_direc);
			if (LOOK_nexti>-1)
				if (q5[LOOK_nexti].cyl==LOOK_head)
				{ 
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						LOOK_nexti = get_cyl_nexti(q5,LOOK_last, LOOK_nexti);
					#endif
					NORWAIT(LOOK,q5);
					LOOK_insrv = SRVTIME(LOOK,q5); update_progress_bar(); goto again5;
				}
				else
					LOOK_direc=-LOOK_direc;
			if (LOOK_cleft) 
			{	LOOK_direc=-LOOK_direc; LOOK_head+=LOOK_direc; 
				
				#ifdef USE_SETTLE
					// If the head has just arrived at the destination cylinder then add "settle time"
					if (LOOK_nexti==-1) LOOK_nexti = get_SCAN_nexti(q5,LOOK_last,LOOK_head,LOOK_direc);
					if (q5[LOOK_nexti].cyl==LOOK_head)		// no need for (LOOK_nexti>-1)
					{	LOOK_seek+=SETTLE_CYLSEEK; LOOK_headmv=SETTLE_CYLSEEK_1; }
					else
				#endif
					{	LOOK_seek+=CYLSEEK; LOOK_headmv=CYLSEEK_1; }
			}
		}
		//LOOK_wait = LOOK_wait+LOOK_cleft;


		//C-LOOK
again6:	if (CLOOK_headmv)
			CLOOK_headmv--;
		else if (CLOOK_insrv)
		{	RWAIT(CLOOK,q6);
			CLOOK_insrv--; if (!CLOOK_insrv) { q6[CLOOK_nexti]=q6[CLOOK_last]; CLOOK_last--; CLOOK_cleft--; CLOOK_nexti = -1; } //else RWAIT(CLOOK);
		} 
		else
		{	if (CLOOK_direc==1)
			{	if (CLOOK_nexti==-1) CLOOK_nexti = get_SCAN_nexti(q6,CLOOK_last,CLOOK_head,CLOOK_direc);
				if (CLOOK_nexti>-1)
					if (q6[CLOOK_nexti].cyl==CLOOK_head)
					{
						#ifdef RLATENCY
							// If rlat_method=2, this will search for the nearest request in current cylinder.
							CLOOK_nexti = get_cyl_nexti(q6,CLOOK_last, CLOOK_nexti);
						#endif
						NORWAIT(CLOOK,q6);
						CLOOK_insrv = SRVTIME(CLOOK,q6); update_progress_bar(); goto again6; 
					}
					else
						CLOOK_direc=-CLOOK_direc;
				if (CLOOK_cleft) 
				{	CLOOK_direc=-CLOOK_direc; CLOOK_head+=CLOOK_direc; 
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (CLOOK_nexti>-1 && q6[CLOOK_nexti].cyl==CLOOK_head)
						{	CLOOK_seek+=SETTLE_CYLSEEK; CLOOK_headmv=SETTLE_CYLSEEK_1; }
						else
					#endif
						{	CLOOK_seek+=CYLSEEK; CLOOK_headmv=CYLSEEK_1; }
				}
			}
			else 
			{	if (CLOOK_nexti==-1) CLOOK_nexti = get_SCAN_nexti(q6,CLOOK_last,STARTCYL,1);
				if (CLOOK_nexti==-1 || q6[CLOOK_nexti].cyl>=CLOOK_head) { CLOOK_direc=1; goto again6; }
				else 
				{	CLOOK_head--;
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (q6[CLOOK_nexti].cyl==CLOOK_head)	// no need for (CLOOK_nexti>-1)
						{	CLOOK_seek+=SETTLE_CYLSEEK; CLOOK_headmv=SETTLE_CYLSEEK_1; }
						else
					#endif
						{	CLOOK_seek+=CYLSEEK; CLOOK_headmv=CYLSEEK_1; }
				}
			}
		}
		//CLOOK_wait = CLOOK_wait+CLOOK_cleft;


		#ifdef USE_SLOOK
		// S-LOOK
again7:	if (SLOOK_headmv)
			SLOOK_headmv--;
		else if (SLOOK_insrv)
		{	RWAIT(SLOOK,q7); SLOOK_insrv--;
			if (!SLOOK_insrv) 
			{	q7[SLOOK_nexti]=q7[SLOOK_last]; SLOOK_last--; SLOOK_cleft--; a=0;
				if (SLOOK_nexti==SLOOK_mini) { SLOOK_mini=get_SCAN_nexti(q7,SLOOK_last,STARTCYL,1); a=1; if (SLOOK_mini==-1) SLOOK_mini=total_requests; }
				else if (SLOOK_mini==(SLOOK_last+1)) SLOOK_mini=SLOOK_nexti;
				if (SLOOK_nexti==SLOOK_maxi) { SLOOK_maxi=get_SCAN_nexti(q7,SLOOK_last,MAXCYLINDER,-1); a=2; if (SLOOK_maxi==-1) SLOOK_maxi=total_requests+1; }
				else if (SLOOK_maxi==(SLOOK_last+1)) SLOOK_maxi=SLOOK_nexti;
				
				if (SLOOK_resmeth==1)
				{	if (a) { SLOOK_waist=SLOOK_wtconst; /*printf("|");*/ }
				}
				else if (SLOOK_resmeth==2 && a!=SLOOK_flag1) { SLOOK_waist=SLOOK_wtconst; SLOOK_flag1=a; /*printf("|");*/ } 
				SLOOK_nexti = -1;
			} //else RWAIT(SLOOK);
		} 
		else
		{	if (SLOOK_nexti==-1) SLOOK_nexti = get_SCAN_nexti(q7,SLOOK_last,SLOOK_head,SLOOK_direc);
			if (SLOOK_nexti>-1)
				if (q7[SLOOK_nexti].cyl==SLOOK_head)
				{
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						SLOOK_nexti = get_cyl_nexti(q7,SLOOK_last, SLOOK_nexti);
					#endif
					NORWAIT(SLOOK,q7);
					SLOOK_insrv = SRVTIME(SLOOK,q7); update_progress_bar(); /*printf("A");*/ goto again7; 
				}
				else
					SLOOK_direc=-SLOOK_direc;
			if (SLOOK_cleft)
			{	SLOOK_direc = -SLOOK_direc; SLOOK_head+=SLOOK_direc; 
				
				//if (SLOOK_direc==1) printf("+"); else printf("-");
				
				#ifdef USE_SETTLE
					// If the head has just arrived at the destination cylinder then add "settle time"
					if (SLOOK_nexti==-1) SLOOK_nexti = get_SCAN_nexti(q7,SLOOK_last,SLOOK_head,SLOOK_direc);
					if (q7[SLOOK_nexti].cyl==SLOOK_head)
					{	SLOOK_seek+=SETTLE_CYLSEEK; SLOOK_headmv=SETTLE_CYLSEEK_1; }
					else
				#endif
					{	SLOOK_seek+=CYLSEEK; SLOOK_headmv=CYLSEEK_1; }
				
				SLOOK_waist+=SLOOK_wtfact;
				
				// Handle over flows
				if (SLOOK_waist+MAXCYLINDER+10<0) SLOOK_waist = INFINITY;	// 10 for safty, it shoulbe be 1 only
			}
			// if method 3, reset waisted time when queue is empty 
			else if (SLOOK_resmeth==3) { SLOOK_waist=SLOOK_wtconst; /*printf("|");*/ }

		}
		//SLOOK_wait = SLOOK_wait+SLOOK_cleft;
		#endif


#ifdef USE_SATF
		//SATF
again9:	if (SATF_headmv)
			SATF_headmv--;
		else if (SATF_insrv)
		{	//if ((SATF_insrv==BLOCKTIME) && cu_bltime!=q9[SATF_nexti].bltime) printf("*");
			RWAIT(SATF,q9); SATF_insrv--; if (!SATF_insrv) { q9[SATF_nexti]=q9[SATF_last]; SATF_last--; SATF_cleft--; SATF_nexti = -1; } //else RWAIT(SATF);
		} 
		else 
		{	if (SATF_nexti==-1) SATF_nexti = get_SATF_nexti_rlat(q9,SATF_last,SATF_head);
			if (SATF_nexti>-1)
			{	n = q9[SATF_nexti].cyl-SATF_head;
				if (n!=0)
				{	n=n/abs(n); SATF_head = SATF_head + n; 
					#ifdef USE_SETTLE
						// If the head has just arrived at the destination cylinder then add "settle time"
						if (q9[SATF_nexti].cyl==SATF_head)
						{	SATF_seek+=SETTLE_CYLSEEK; SATF_headmv=SETTLE_CYLSEEK_1; }
						else
					#endif
						{ SATF_seek+=CYLSEEK; SATF_headmv=CYLSEEK_1;
						}
				} else 
				{	NORWAIT(SATF,q9);
					SATF_insrv = SRVTIME(SATF,q9); update_progress_bar(); goto again9;
				}
			}
		}
		//SATF_wait = SATF_wait+SATF_cleft;
#endif


		#ifdef RLATENCY
			// Rotate disk by a time unit. If block zero is reached reset time to 0
			cu_bltime++; if (cu_bltime>=BPTTIME) cu_bltime=0;
		#endif

		// Increase Simulation Timer
		cu_time++;

		// Check for pending requests for next loop 
		all_cleft = SSTF_cleft || FCFS_nexti<=last || SCAN_cleft || CSCAN_cleft || LOOK_cleft || CLOOK_cleft || SLOOK_cleft || BOPATH_cleft ON_SATF(|| SATF_cleft);

	} // Main Simulation Loop

	if (FCFS_nexti==total_requests+10) FCFS_wait=0;

	a=(rcount-1)*BLOCKTIME;
	if (sstream!=NULL)
	{	if (scolumn!=NULL) fprintf(sstream,scolumn); fprintf(sstream,",");
		fprintf(sstream,"%I64d,",FCFS_seek); 
		fprintf(sstream,"%I64d,",SSTF_seek); 
		fprintf(sstream,"%I64d,",SCAN_seek); 
		fprintf(sstream,"%I64d,",CSCAN_seek);
		fprintf(sstream,"%I64d,",LOOK_seek);
		fprintf(sstream,"%I64d,",CLOOK_seek);
		fprintf(sstream,"%I64d,",SLOOK_seek);
		fprintf(sstream,"%I64d,",BOPATH_seek);
		#ifdef USE_SATF
			fprintf(sstream,"%I64d\n",SATF_seek);
		#else
			fprintf(sstream,"0\n");
		#endif
		fclose(sstream);
	}
	if (wstream!=NULL)
	{	if (wcolumn!=NULL) fprintf(wstream,wcolumn); fprintf(wstream,",");
		fprintf(wstream,"%g,",(double) FCFS_wait / rcount);
		fprintf(wstream,"%g,",(double) SSTF_wait / rcount); 
		fprintf(wstream,"%g,",(double) SCAN_wait / rcount); 
		fprintf(wstream,"%g,",(double) CSCAN_wait / rcount);
		fprintf(wstream,"%g,",(double) LOOK_wait / rcount); 
		fprintf(wstream,"%g,",(double) CLOOK_wait / rcount); 
		fprintf(wstream,"%g,",(double) SLOOK_wait / rcount);
		fprintf(wstream,"%g,",(double) BOPATH_wait / rcount);
		#ifdef USE_SATF
			fprintf(wstream,"%g\n",(double) SATF_wait / rcount);
		#else
			fprintf(wstream,"0\n");
		#endif
		fclose(wstream);
	}
	if (xwstream!=NULL)
	{	if (xwcolumn!=NULL) fprintf(xwstream,xwcolumn); fprintf(xwstream,",");
		fprintf(xwstream,"%lu,",FCFS_maxwait);
		fprintf(xwstream,"%lu,",SSTF_maxwait); 
		fprintf(xwstream,"%lu,",SCAN_maxwait); 
		fprintf(xwstream,"%lu,",CSCAN_maxwait);
		fprintf(xwstream,"%lu,",LOOK_maxwait); 
		fprintf(xwstream,"%lu,",CLOOK_maxwait); 
		fprintf(xwstream,"%lu,",SLOOK_maxwait);
		fprintf(xwstream,"%lu,",BOPATH_maxwait);
		#ifdef USE_SATF
			fprintf(xwstream,"%lu\n",SATF_maxwait);
		#else
			fprintf(xwstream,"0\n");
		#endif
		fclose(xwstream);
	}
	
	if (sstream==NULL && wstream==NULL && xwstream==NULL)
	{	printf("\nFCFS total seek = %I64d", FCFS_seek);
		printf("\nFCFS total seek (including block time) = %I64d", FCFS_seek + a);
		printf("\nFCFS maximum waiting time = %lu", FCFS_maxwait);
		printf("\nFCFS average waiting time = %f\n", (double) FCFS_wait / rcount);

		printf("\nSSTF total seek = %I64d", SSTF_seek);
		printf("\nSSTF total seek (including block time) = %I64d", SSTF_seek + a);
		printf("\nSSTF maximum waiting time = %lu", SSTF_maxwait);
		printf("\nSSTF average waiting time = %f\n", (double) SSTF_wait / rcount);

		printf("\nSCAN total seek = %I64d", SCAN_seek);
		printf("\nSCAN total seek (including block time) = %I64d", SCAN_seek + a);
		printf("\nSCAN maximum waiting time = %lu", SCAN_maxwait);
		printf("\nSCAN average waiting time = %f\n", (double) SCAN_wait / rcount);

		printf("\nC-SCAN total seek = %I64d", CSCAN_seek);
		printf("\nC-SCAN total seek (including block time) = %I64d", CSCAN_seek + a);
		printf("\nC-SCAN maximum waiting time = %lu", CSCAN_maxwait);
		printf("\nC-SCAN average waiting time = %f\n", (double) CSCAN_wait / rcount);

		printf("\nLOOK total seek = %I64d", LOOK_seek);
		printf("\nLOOK total seek (including block time) = %I64d", LOOK_seek + a);
		printf("\nLOOK maximum waiting time = %lu", LOOK_maxwait);
		printf("\nLOOK average waiting time = %f\n", (double) LOOK_wait / rcount);

		printf("\nC-LOOK total seek = %I64d", CLOOK_seek);
		printf("\nC-LOOK total seek (including block time) = %I64d", CLOOK_seek + a);
		printf("\nC-LOOK maximum waiting time = %lu", CLOOK_maxwait);
		printf("\nC-LOOK average waiting time = %f\n", (double) CLOOK_wait / rcount);
	
		printf("\nS-LOOK total seek = %I64d", SLOOK_seek);
		printf("\nS-LOOK total seek (including block time) = %I64d", SLOOK_seek + a);
		printf("\nS-LOOK maximum waiting time = %lu", SLOOK_maxwait);
		printf("\nS-LOOK average waiting time = %f\n", (double) SLOOK_wait / rcount);

		printf("\nBO-PATH total seek = %I64d", BOPATH_seek);
		printf("\nBO-PATH total seek (including block time) = %I64d", BOPATH_seek + a);
		printf("\nBO-PATH maximum waiting time = %lu", BOPATH_maxwait);
		printf("\nBO-PATH average waiting time = %f\n", (double) BOPATH_wait / rcount);

		#ifdef USE_SATF
			printf("\nSATF total seek = %I64d", SATF_seek);
			printf("\nSATF total seek (including block time) = %I64d", SATF_seek + a);
			printf("\nSATF maximum waiting time = %lu", SATF_maxwait);
			printf("\nSATF average waiting time = %f\n", (double) SATF_wait / rcount);
		#else
			printf("\nYou should #define USE_SATF to get SATF results.\n");
		#endif
	}

	/* printf("\n");
	for (n=0; n<=last; n++)
		printf(" %d",q[n].cyl);
	printf("\n");
	free(q2); q2=malloc(sizeof(int)*(total_requests));	
	printf("Optimal average waiting time = %g\n",(double) ASTAR_opt(q,last,head,q2,&a,BLOCKTIME,CYLSEEK,1) / rcount);
	printf("Total seek = %d\n",a);
	for (n=0; n<=last; n++)
		printf(" %d",q[q2[n]].cyl); */

EXIT_MAIN:
	if (caller_hwnd != -1) PostMessage((HWND) caller_hwnd,WM_PROGRESS,chunk_count,0);

	// Cleaning
	_fcloseall();
	free(q); free(q2); free(q3); free(q4); 
	free(q5); free(q6); free(q7); free(q8); 
	ON_SATF(free(q9)); free(qopti); free(qoptinv); free(qm);
}


/*

int insort(qreq *q, int last, int val)
{	int i=0,n;
	while (val>q[i].cyl && i<=last) i++;
	for (n=last+1; n>i; n--) q[n].cyl=q[n-1].cyl;
	q[i]=val; return i;
}
*/
