Attribute VB_Name = "ArbitraryPrecision"
#If APM Then
Option Explicit
Declare Function GetTickCount Lib "kernel32" () As Long
'////////////////////////////////////////////////////////////////////////////////
'///                                                                          ///
'///                                Encryption                                ///
'///                           modArbitraryPrecision                          ///
'///                                                                          ///
'////////////////////////////////////////////////////////////////////////////////
'///                                                                          ///
'///                                  Author                                  ///
'///                                                                          ///
'/// Name:      Andrew Mallner                                                ///
'/// Email:     darkforcesjedi@ec.rr.com                                      ///
'/// Copyright: (c) 2004 Andrew Mallner, all rights reserved                  ///
'///            this code may be used for educational purposes only and may   ///
'///            not be redistributed                                          ///
'///                                                                          ///
'////////////////////////////////////////////////////////////////////////////////
'///                                                                          ///
'///                            Version Information                           ///
'///                                                                          ///
'/// Version:      0.1.0                                                      ///
'/// Date Revised: 06-Jul-2004                                                ///
'///                                                                          ///
'////////////////////////////////////////////////////////////////////////////////

' Arbitrary precision base 65536 arithmetic
' Lowest digit is in the lowest index

'note that integers are signed, so if the high-order bit is 1 then the
'number is the unsigned value - 65536
'
'Digit order:
'0, 1, 2, 3, ... 32767, -32768, -32767, ... -1

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
    dest As Any, source As Any, ByVal bytes As Long)

'I'm too lazy to deal with signing
'instead the sign is stored in a separate boolean

Public Type HugeInt
    Digits() As Integer
    IsNegative As Boolean
End Type

Public Enum Comparisons
    Equal = 0
    Greater = 1
    Less = 2
End Enum

Public PrimeInit As Boolean, Bases() As Integer
Public FactorOut() As HugeInt

Public Function Add(a As HugeInt, b As HugeInt) As HugeInt
    'Debug.Print "Add"
    'Adds b to a: Add = a + b
    If Not (a.IsNegative Xor b.IsNegative) Then
        'Either both are negative or both
        'are positive
        Add.Digits = prvAdd(a, b).Digits
        Add.IsNegative = a.IsNegative
    ElseIf a.IsNegative Then
        If Compare(a, b) = Greater Then
            Add.Digits = prvSubtract(a, b).Digits
            Add.IsNegative = True
        Else
            Add.Digits = prvSubtract(b, a).Digits
        End If
    Else
        'b is negative, a is not
        If Compare(a, b) = Less Then
            Add.Digits = prvSubtract(b, a).Digits
            Add.IsNegative = True
        Else
            Add.Digits = prvSubtract(a, b).Digits
        End If
    End If
End Function

Public Sub Append(This() As HugeInt, ToThis() As HugeInt)
    Dim i As Long, j As Long, k As Long
    
    j = -1
    k = -1
    
    On Error Resume Next
    j = UBound(ToThis)
    k = UBound(This)
    On Error GoTo 0
    
    If k < 0 Then Exit Sub
    
    ReDim Preserve ToThis(j + k + 1)
    
    For i = 0 To k
        ToThis(i + j + 1) = Equals(This(i))
    Next
End Sub

Public Function aShift(a As HugeInt, Shift As Long) As HugeInt

    Dim iBuff() As Integer
    ReDim iBuff(UBound(a.Digits) + Shift)
    
    If Shift > 0 Then
        CopyMemory iBuff(Shift), a.Digits(0), (UBound(a.Digits) + 1) * 2
    ElseIf Shift < 0 Then
        CopyMemory iBuff(0), a.Digits(-Shift), (UBound(iBuff) + 1) * 2
    Else
        CopyMemory iBuff(0), a.Digits(0), (UBound(a.Digits) + 1) * 2
    End If
    
    aShift.Digits = iBuff
    Erase iBuff
    aShift.IsNegative = a.IsNegative

End Function

Public Function Compare(This As HugeInt, ToThis As HugeInt, _
    Optional IgnoreSigns As Boolean = True) As Comparisons
    'Debug.Print "Compare"
    Dim ua            As Long
    Dim ub            As Long
    Dim lPlaceHolder  As Long
    Dim lPlaceHolder2 As Long
    
    If Not IgnoreSigns Then
        If This.IsNegative Xor ToThis.IsNegative Then
            If This.IsNegative Then
                Compare = Less
            Else
                Compare = Greater
            End If
            Exit Function
        End If
    End If
    
    ua = UBound(This.Digits)
    ub = UBound(ToThis.Digits)
    
    If ua > ub Then
        Compare = Greater
        Exit Function
    ElseIf ub > ua Then
        Compare = Less
        Exit Function
    End If
    
    For ub = ua To 0 Step -1
        lPlaceHolder = This.Digits(ub) And &HFFFF&
        lPlaceHolder2 = ToThis.Digits(ub) And &HFFFF&
        
        If Not (lPlaceHolder = lPlaceHolder2) Then
            If lPlaceHolder > lPlaceHolder2 Then
                Compare = Greater
                'Debug.Print "End Compare - Greater"
                Exit Function
            Else
                Compare = Less
                'Debug.Print "End Compare - Less"
                Exit Function
            End If
        End If
    Next
    'Debug.Print "End Compare - Equal"
End Function

Public Function DecToHugeInt(sDigits As String, Optional IsNeg As Boolean = False) As HugeInt

    Dim x() As Byte, i As Long, p As HugeInt, Ten As HugeInt, b As HugeInt
    
    Ten = IntOut(10)
    p = IntOut(0)
    b = One
    x = StrConv(sDigits, vbFromUnicode)
    
    For i = UBound(x) To 0 Step -1
        p = Add(p, Multiply(b, IntOut(x(i) - 48)))
        b = Multiply(b, Ten)
    Next
    DecToHugeInt = Equals(p)
    
    Erase p.Digits
    Erase b.Digits
    Erase Ten.Digits
    Erase x
End Function

Public Function Divide(Numerator As HugeInt, Denominator As HugeInt) As HugeInt
    'Debug.Print "Divide"
    Dim r As HugeInt
    Divide = prvDivide(Numerator, Denominator, r)
    Erase r.Digits
    Divide.IsNegative = Numerator.IsNegative Xor Denominator.IsNegative
End Function

Public Sub DivMod(Numerator As HugeInt, Denominator As HugeInt, _
    ByRef Quotient As HugeInt, ByRef Remainder As HugeInt)
    Quotient = prvDivide(Numerator, Denominator, Remainder)
End Sub

Public Sub DumpFactors(n As HugeInt, Optional DumpInHex As Boolean = True)

    Factor n
    
    Dim i As Long
    
    If DumpInHex Then
        For i = 0 To UBound(FactorOut)
            DumpNum FactorOut(i)
        Next
    Else
        For i = 0 To UBound(FactorOut)
            DumpNumDec FactorOut(i)
        Next
    End If

End Sub

Public Sub DumpNum(c As HugeInt)
    Debug.Print "   ";
    Dim i As Long, j As Long
    j = UBound(c.Digits)
    For i = j To 0 Step -1
        If i = j Then Debug.Print Hex(c.Digits(i)); Else Debug.Print Right("000" & Hex(c.Digits(i)), 4);
    Next
    Debug.Print
End Sub

Public Sub DumpNumDec(a As HugeInt) ', Optional ByVal ChunkSize As Long = 100)
    Dim i As Long, u As Long
    Dim b() As Variant
    
    b = HugeIntToDec(a)
    u = UBound(b)
    Debug.Print "   ";
    For i = UBound(b) To 0 Step -1
        Debug.Print IIf(i < UBound(b), Right$("0000000000000000000000000000" & CStr(b(i)), 28), CStr(b(i)));
        'If u - i + IIf(u = UBound(b), 1, 0) = ChunkSize Then u = i: Debug.Print "": Debug.Print "   ";
    Next
    Debug.Print ""
    Erase b
End Sub

Public Sub DumpNumToFile(c As HugeInt, filename As String)
    Open filename For Append As #1
    Dim i As Long, j As Long
    j = UBound(c.Digits)
    For i = j To 0 Step -1
        If i = j Then Print #1, Hex(c.Digits(i)); Else Print #1, Right("000" & Hex(c.Digits(i)), 4);
    Next
    Print #1, ""
    Close #1
End Sub

Public Function Equals(a As HugeInt) As HugeInt
    Equals.Digits = a.Digits
    Equals.IsNegative = a.IsNegative
End Function

Public Function f(ByVal l As Long) As Integer
    If l And &H8000 Then
        f = l - 65536
    Else
        f = l
    End If
End Function

Public Sub Factor(n As HugeInt, Optional ByVal ArgumentMightBePrime As Boolean = True, Optional ByVal ClearFactorOut As Boolean = True)
    If ClearFactorOut Then Erase FactorOut
    Randomize
    Dim hBuff() As HugeInt
    
    Dim lCount As Long
    
    Dim x() As HugeInt, i As Long, j As Long, k As Long, factorh As Boolean
    Dim q As HugeInt, r As HugeInt, s As HugeInt, g As HugeInt, h As HugeInt, p As HugeInt
    
    ReDim x(Int(256 * UBound(n.Digits)))
    
    If ArgumentMightBePrime Then
        If IsPrime(n) Then
            ReDim hBuff(0)
            hBuff(0) = Equals(n)
            Append hBuff, FactorOut
            GoTo cleanup
        End If
    End If
    
    If (n.Digits(0) And 1) = 0 Then
        'divisible by 2
        g = IntOut(2)
        ReDim hBuff(0)
        hBuff(0) = Equals(g)
        h = Divide(n, g)
        Append hBuff, FactorOut
        Factor h, , False
        GoTo cleanup
    End If
    
    If UBound(n.Digits) = 0 Then
        'Running out of stack space when factoring small numbers
        k = n.Digits(0) And &HFFFF&
        Dim fa() As Double
        fa = sFactor(k)
        ReDim hBuff(0)
        For i = 0 To UBound(fa)
            hBuff(0) = IntOut(f(fa(i)))
            Append hBuff, FactorOut
        Next
        Erase fa
        GoTo cleanup
    End If
    
    j = 1
    lCount = -1
    
    x(0) = Add(Rand(1), One)
    q = IntOut(Int(Rnd * 3) + 1)
    r = Add(Rand(1), One)
    's = Rand(Int(Rnd * 3))
    'p = IntOut(Int(Rnd * 4) + 1)
    
    For i = 1 To UBound(x)
        'x(i) = Subtract(Multiply(q, Power(x(i - 1), p)), Multiply(r, x(i - 1)))
        'x(i) = Modulo(x(i), n)
        'x(i) = Modulo(Add(x(i), s), n)
        
        x(i) = Subtract(prvModExp(x(i - 1), q, n), r)
        x(i) = Modulo(x(i), n)
        
        'x(i+1) = q*x(i)^2 - r*x(i) + s  (mod n)
        
        
        
        If i = 2 * j Then
            If IsEqual(x(i), x(j)) Then Exit For
            
            g = GCD(Subtract(x(i), x(j)), n)
            
            If IsGreater(g, One) And Not IsEqual(g, n) Then
                Factor g, , False
                h = Divide(n, g)
                factorh = True
                Exit For
            End If
            
            j = j + 1
        End If
    Next
    
    If factorh Then
        Factor h, , False
    Else
        Factor n, False, False
    End If
    
cleanup:
    Erase x
    Erase g.Digits
    Erase h.Digits
    Erase q.Digits
    Erase r.Digits
    Erase s.Digits
    Erase p.Digits
    
End Sub

Public Function GCD(n1 As HugeInt, n2 As HugeInt) As HugeInt

    Dim a As HugeInt
    Dim b As HugeInt
    Dim c As Comparisons
    a.Digits = n1.Digits
    b.Digits = n2.Digits
    
    Do Until IsZero(a) Or IsZero(b)
        c = Compare(a, b)
        If c = Greater Then
            a = Modulo(a, b)
        ElseIf c = Less Then
            b = Modulo(b, a)
        Else
            GCD.Digits = a.Digits
            Exit Function
        End If
    Loop
    
    If IsZero(a) Then
        GCD.Digits = b.Digits
    Else
        GCD.Digits = a.Digits
    End If

End Function

Public Function HugeIntToDec(a As HugeInt) As Variant() 'vbDecimal
    Dim k As Integer
    Dim r As HugeInt
    Dim remain As HugeInt
    Dim quotie As HugeInt
    Dim Ten As HugeInt
    Dim lPlace As Long
    Dim vBuff() As Variant 'max Decimal=79,228,162,514,264,337,593,543,950,335
    Dim a1, a2, a3, a4, a5
    
    a1 = CDec(65536)
    a2 = a1 * a1
    a3 = a2 * a1
    a4 = a3 * a1
    a5 = a4 * a1
    
    
    r = Equals(a)
    Ten = DecToHugeInt("10000000000000000000000000000")
    
    ReDim vBuff(UBound(r.Digits))
    
    Do Until IsZero(r)
        DivMod r, Ten, quotie, remain
        r = Equals(quotie)
        
        vBuff(lPlace) = (remain.Digits(0) And &HFFFF&)
        k = UBound(remain.Digits)
        
        If k > 0 Then
            vBuff(lPlace) = vBuff(lPlace) + (CDec(remain.Digits(1) And &HFFFF&) * a1)
        End If
        If k > 1 Then
            vBuff(lPlace) = vBuff(lPlace) + (CDec(remain.Digits(2) And &HFFFF&) * a2)
        End If
        If k > 2 Then
            vBuff(lPlace) = vBuff(lPlace) + (CDec(remain.Digits(3) And &HFFFF&) * a3)
        End If
        If k > 3 Then
            vBuff(lPlace) = vBuff(lPlace) + (CDec(remain.Digits(4) And &HFFFF&) * a4)
        End If
        If k > 4 Then
            vBuff(lPlace) = vBuff(lPlace) + (CDec(remain.Digits(5) And &HFFFF&) * a5)
        End If
        
        r.Digits = TrimLeft(r.Digits)
        lPlace = lPlace + 1
        'If (lPlace Mod 10 = 0) Then
        '    Debug.Print CStr(lPlace); " of "; CStr(UBound(iBuff))
        '    DoEvents
        'End If
    Loop
    
    Dim i As Long
    For i = UBound(vBuff) To 0 Step -1
        If vBuff(i) <> 0 Then
            ReDim Preserve vBuff(i)
            Exit For
        End If
    Next
    
    HugeIntToDec = vBuff
    
    Erase r.Digits
    Erase remain.Digits
    Erase quotie.Digits
    Erase Ten.Digits
    Erase vBuff
End Function

Public Function HugeIntToDecString(a As HugeInt) As String
    Dim x As Variant, o As String, s As String, i As Long, offset As Long
    
    x = HugeIntToDec(a)
    o = CStr(x(UBound(x)))
    
    If UBound(x) > 0 Then
        offset = LenB(o)
        o = o & String(28 * UBound(x), "0")
        
        For i = UBound(x) - 1 To 0 Step -1
            s = CStr(x(i))
            offset = offset + LenB(Space(28)) - LenB(s)
            CopyMemory ByVal StrPtr(o) + offset, ByVal StrPtr(s), LenB(s)
            offset = offset + LenB(s)
        Next
    End If
    Erase x
    HugeIntToDecString = o
End Function

Public Function IntOut(intIn As Integer, _
    Optional IsNeg As Boolean = False) As HugeInt
    ReDim IntOut.Digits(0)
    IntOut.Digits(0) = intIn
    IntOut.IsNegative = IsNeg
End Function

Public Function IsEqual(a As HugeInt, b As HugeInt) As Boolean
    'Debug.Print "IsEqual"
    IsEqual = (Compare(a, b) = Equal)
End Function

Public Function IsGreater(IsThisGreater As HugeInt, _
    ThanThis As HugeInt) As Boolean
    'Debug.Print "IsGreater"
    IsGreater = (Compare(IsThisGreater, ThanThis, False) = Greater)
End Function

Public Function IsGreaterOrEqual(IsThisGreater As HugeInt, _
    ThanThis As HugeInt) As Boolean
    'Not < is the same as >=
        'Debug.Print "IsGreaterOrEqual"
    IsGreaterOrEqual = Not (Compare(IsThisGreater, ThanThis, False) = Less)
End Function

Public Function IsLess(IsThisLess As HugeInt, ThanThis As HugeInt) As Boolean
    'Debug.Print "IsLess"
    IsLess = (Compare(IsThisLess, ThanThis, False) = Less)
End Function

Public Function IsLessOrEqual(IsThisLess As HugeInt, _
    ThanThis As HugeInt) As Boolean
    'Debug.Print "IsLEssOrEqual"
    IsLessOrEqual = Not (Compare(IsThisLess, ThanThis, False) = Greater)
End Function

Public Function IsOne(a As HugeInt) As Boolean

    If UBound(a.Digits) = 0 And a.Digits(0) = 1 Then IsOne = True

End Function

Public Function IsPowerOf2(n As HugeInt) As Boolean
    
    If Log(n.Digits(UBound(n.Digits))) / Log(2) = _
       Int(Log(n.Digits(UBound(n.Digits))) / Log(2)) Then
        
        Dim i As Long
        For i = UBound(n.Digits) - 1 To 0 Step -1
            If n.Digits(i) <> 0 Then Exit Function
        Next
        IsPowerOf2 = True
    End If
    
End Function

Public Function IsPrime(a As HugeInt) As Boolean
    'This uses a slow probablistic method to test for
    'primality.
    '
    'If GCD(a,x) = 1 and x^(n-1) mod n = 1 or -1 then
    '  n fails Miller's test for at least half of the
    '  bases 1 < x < n
    '
    'All odd numbers are coprime to 2, i.e. if n is odd
    'then GCD(n,2)=1. A number is odd if the last bit is
    '1, i.e. (a.digits(0) And 1) = 1
    '
    'Thus after 100 passed tests we are
    ' 99.999999999999999999999999999921% positive that a
    ' is prime
    '
    'a is assumed positive
    
    If IsOne(a) Then Exit Function
    
    If Not PrimeInit Then InitPrimeBases
    Dim w As Integer
    If UBound(a.Digits) = 0 Then
        For w = 0 To UBound(Bases)
            If Bases(w) = a.Digits(0) Then IsPrime = True: Exit Function
            If (a.Digits(0) And &HFFFF&) < Bases(w) + 1 Then Exit Function
        Next
    End If
    
    If (a.Digits(0) And 1) = 0 Then Exit Function
    
    Dim Base As HugeInt, i As Byte, aminusone As HugeInt, max As Byte, _
        result As HugeInt
    w = 0
    aminusone = prvSubtract(a, One)
    Base = IntOut(Bases(w))
    
    If IsLessOrEqual(a, IntOut(-1)) Then
        max = 10
    Else
        'Allows ~1 in 1,050,000 composites to pass
        max = 10
    End If
    
    For i = 1 To max
        result = prvModExp(Base, aminusone, a)
        If Not (IsOne(result) Or IsEqual(result, aminusone)) Then
            'failed the test
            GoTo cleanup
        End If
        w = w + 1
        Base = IntOut(Bases(w))
    Next
    IsPrime = True
cleanup:

    Erase Base.Digits
    Erase aminusone.Digits
    Erase result.Digits
    
End Function

Public Function IsZero(a As HugeInt) As Boolean

    If UBound(a.Digits) = 0 And a.Digits(0) = 0 Then IsZero = True

End Function

Public Function ModularInverse(a As HugeInt, Modulus As HugeInt) As HugeInt
    ' Reutrns 0 if there is no inverse
    ' Returns the inverse if it exists
    '
    ' If (a,Modulus) <> 1 then the inverse is undefined
    ' however the function does not make this check to save
    ' CPU cycles. This is not a problem if the modulus is prime
    ' and a is less than the modulus
    '
    ' Modular inverse of 15 mod 26:
    '
    ' 26  1  0
    ' 15  0  1 'divide 26 by 15 and subtract 26\15 * this row from row 1
    ' 11  1 -1 'divide 15 by 11 and subtract 15\11 * this row from row 2
    '  4 -1  2 ' ...
    '  3  3 -5
    '  1 -4  7 'when column 1 = 1, column 3 is the modular inverse
    
    a = Modulo(a, Modulus)
    
    Dim a1 As HugeInt, a2 As HugeInt, a3 As HugeInt
    Dim b1 As HugeInt, b2 As HugeInt, b3 As HugeInt
    Dim c1 As HugeInt, c2 As HugeInt, c3 As HugeInt
    Dim q As HugeInt
    
    a1.Digits = Modulus.Digits
    a2 = One
    a3 = Zero
    
    b1.Digits = a.Digits
    b2 = Zero
    b3 = One
    
    Do Until IsOne(a1) Or IsZero(b1) 'if it reaches zero then there is no inverse
        q = prvDivide(a1, b1, c1)
        c2 = Subtract(a2, Multiply(q, b2))
        c3 = Subtract(a3, Multiply(q, b3))
        
        a1 = Equals(b1)
        a2 = Equals(b2)
        a3 = Equals(b3)
        
        b1 = Equals(c1)
        b2 = Equals(c2)
        b3 = Equals(c3)
    Loop
    
    If IsOne(a1) Then
        ModularInverse = Modulo(a3, Modulus)
    Else
        ReDim ModularInverse.Digits(0)
    End If
    
    Erase a1.Digits
    Erase a2.Digits
    Erase a3.Digits
    Erase b1.Digits
    Erase b2.Digits
    Erase b3.Digits
    Erase c1.Digits
    Erase c2.Digits
    Erase c3.Digits
    Erase q.Digits
End Function

Public Function Modulo(Number As HugeInt, Modulus As HugeInt) As HugeInt
    'Debug.Print "Modulo"
    Dim r As HugeInt
    prvDivide Number, Modulus, r
    If Number.IsNegative Then
        ' -8 / -3 = 3 with remainder 1 = 2 remainder -2
        ' -3 * 3 = -9... + 1 = -8
        '
        ' 8 / -3 = -2 remainder 2
        ' 8 / 3  = 2  remainder 2
        ' -8 / 3 = -3 remainder 1 = -2 remainder -2
        '
        ' if the remainder is negative we subtract it from the modulus
        ' use prvSubtract because we already know r < modulus
        r = prvSubtract(Modulus, r)
    End If
    
    Modulo.Digits = r.Digits
End Function

Public Function Multiply(a As HugeInt, b As HugeInt) As HugeInt
    'Debug.Print "Multiply"
    Dim ua As Long
    Dim ub As Long
    Dim iBuff() As Integer
    
    ua = UBound(a.Digits)
    ub = UBound(b.Digits)
    
    'multiply by 0
    If (ua And ub) = 0 Then
        If (a.Digits(0) = 0 And ua = 0) Or (b.Digits(0) = 0 And ub = 0) Then
            ReDim iBuff(0)
            Multiply.Digits = iBuff
            Erase iBuff
            Exit Function
        ElseIf (a.Digits(0) = 1 And ua = 0) Then
            Multiply.Digits = b.Digits
            Multiply.IsNegative = a.IsNegative Xor b.IsNegative
            Exit Function
        ElseIf (b.Digits(0) = 1 And ub = 0) Then
            Multiply.Digits = a.Digits
            Multiply.IsNegative = a.IsNegative Xor b.IsNegative
            Exit Function
        End If
    End If
    
    Multiply.Digits = prvMultiply(a, b).Digits
    Multiply.IsNegative = a.IsNegative Xor b.IsNegative
End Function

Public Function NextPrime(a As HugeInt, Optional ByRef NumsTested As Long) As HugeInt
    Dim p As HugeInt, i As Long, k As Long
    If (a.Digits(0) And 1) = 1 Then
        p = Add(a, IntOut(2))
    Else
        p = Add(a, One)
    End If
    i = 1
    k = GetTickCount
    Do Until IsPrime(p)
        i = i + 1
        'Form1.Label9 = "Testing: " & i
        DoEvents
        p = Add(p, IntOut(2))
    Loop
    NumsTested = i
    Debug.Print GetTickCount - k
    NextPrime.Digits = p.Digits
    Erase p.Digits
End Function

Public Function One() As HugeInt

    ReDim One.Digits(0)
    One.Digits(0) = 1

End Function

Public Function Power(Base As HugeInt, Exponent As HugeInt) As HugeInt
    'Debug.Print "Power"
    
    Power = prvPower(Base, Exponent)
    Power.IsNegative = IIf((Base.IsNegative And (Exponent.Digits( _
        0) And 1 = 0)) Or Not Base.IsNegative, False, True)
End Function

Public Function prvModExp(Base As HugeInt, Power As HugeInt, _
    Modulus As HugeInt) As HugeInt
    'Debug.Print "prvModExp"

    Dim tmpSquare As HugeInt
    Dim Output    As HugeInt
    Dim i         As Long
    Dim j         As Long
    Dim k         As Long
    
    tmpSquare = Modulo(Base, Modulus)
    ReDim Output.Digits(0)
    Output.Digits(0) = 1

'    If UBound(base.Digits) = 0 And base.Digits(0) = 2 Then
'
'        If IsGreater(Power, IntOut(16)) Then
'            tmpSquare = aShift(One, 1)
'            tmpSquare = prvModExp(tmpSquare, Divide(Power, IntOut(16)), modulus)
'
'            Output = prvPower(IntOut(2), Modulo(Power, IntOut(16)))
'
'            Output = Modulo(Multiply(tmpSquare, Output), modulus)
'
'            prvModExp = Equals(Output)
'            GoTo finish
'        End If
'
'    End If

    For i = 0 To UBound(Power.Digits) - 1
        For j = 0 To 15
            If (i Or j) Then tmpSquare = prvSquare(tmpSquare)
            tmpSquare = Modulo(tmpSquare, Modulus)
            If IsOne(tmpSquare) Then GoTo finish
            If Power.Digits(i) And 2 ^ j Then
                Output = Modulo(Multiply(Output, tmpSquare), Modulus)
            End If
        Next
    Next
    
    If Power.Digits(i) And &HFFFF& > 0 Then
        k = Int(Log(Power.Digits(i) And &HFFFF&) / Log(2))
    
        For j = 0 To k
            If (i Or j) Then tmpSquare = Modulo(prvSquare(tmpSquare), Modulus)
            If IsOne(tmpSquare) Then GoTo finish
            If Power.Digits(i) And 2 ^ j Then
                Output = Modulo(Multiply(Output, tmpSquare), Modulus)
            End If
        Next
    End If
finish:
    prvModExp.Digits = Output.Digits
    
    Erase Output.Digits
    Erase tmpSquare.Digits
    'Debug.Print "End prvModExp"
End Function

Public Function Rand(NumDigits As Long) As HugeInt
    Dim i As Long
    Randomize
    
    If NumDigits = 0 Then Rand = Zero: Exit Function
    
    ReDim Rand.Digits(NumDigits - 1)
    For i = 0 To NumDigits - 1
        Rand.Digits(i) = f(Int(Rnd * 65536))
    Next
End Function

Public Function Square(a As HugeInt) As HugeInt
    'Debug.Print "Square"
    Square.Digits = prvSquare(a).Digits
End Function

Public Function Subtract(a As HugeInt, b As HugeInt) As HugeInt
    'Debug.Print "Subtract"
    'Subtracts b from a: Subtract = a - b
    If Not (a.IsNegative Xor b.IsNegative) Then
        'Either both are negative or both
        'are positive
        If Compare(a, b) = Greater Then
            Subtract.Digits = prvSubtract(a, b).Digits
            Subtract.IsNegative = a.IsNegative
        Else
            Subtract.Digits = prvSubtract(b, a).Digits
            Subtract.IsNegative = Not b.IsNegative
        End If
    ElseIf a.IsNegative Then
        Subtract.Digits = prvAdd(a, b).Digits
        Subtract.IsNegative = True
    Else
        'b is negative, a is not
        Subtract.Digits = prvAdd(a, b).Digits
    End If
End Function

Public Function Zero() As HugeInt

    ReDim Zero.Digits(0)

End Function

Private Function ArrayPartition(InputArray() As Integer, StartIndex As Long, _
    EndIndex As Long) As Integer()
    Dim buffer() As Integer
    Dim i        As Long
    
    ReDim buffer(EndIndex - StartIndex)
    CopyMemory buffer(0), InputArray(StartIndex), _
        (EndIndex - StartIndex + 1) * 2
    
    ArrayPartition = buffer
    Erase buffer
End Function

Private Sub InitPrimeBases()
    ReDim Bases(99)
    Bases(0) = 2
    Bases(1) = 3
    Bases(2) = 5
    Bases(3) = 7
    Bases(4) = 11
    Bases(5) = 13
    Bases(6) = 17
    Bases(7) = 19
    Bases(8) = 23
    Bases(9) = 29
    Bases(10) = 31
    Bases(11) = 37
    Bases(12) = 41
    Bases(13) = 43
    Bases(14) = 47
    Bases(15) = 53
    Bases(16) = 59
    Bases(17) = 61
    Bases(18) = 67
    Bases(19) = 71
    Bases(20) = 73
    Bases(21) = 79
    Bases(22) = 83
    Bases(23) = 89
    Bases(24) = 97
    Bases(25) = 101
    Bases(26) = 103
    Bases(27) = 107
    Bases(28) = 109
    Bases(29) = 113
    Bases(30) = 127
    Bases(31) = 131
    Bases(32) = 137
    Bases(33) = 139
    Bases(34) = 149
    Bases(35) = 151
    Bases(36) = 157
    Bases(37) = 163
    Bases(38) = 167
    Bases(39) = 173
    Bases(40) = 179
    Bases(41) = 181
    Bases(42) = 191
    Bases(43) = 193
    Bases(44) = 197
    Bases(45) = 199
    Bases(46) = 211
    Bases(47) = 223
    Bases(48) = 227
    Bases(49) = 229
    Bases(50) = 233
    Bases(51) = 239
    Bases(52) = 241
    Bases(53) = 251
    Bases(54) = 257
    Bases(55) = 263
    Bases(56) = 269
    Bases(57) = 271
    Bases(58) = 277
    Bases(59) = 281
    Bases(60) = 283
    Bases(61) = 293
    Bases(62) = 307
    Bases(63) = 311
    Bases(64) = 313
    Bases(65) = 317
    Bases(66) = 331
    Bases(67) = 337
    Bases(68) = 347
    Bases(69) = 349
    Bases(70) = 353
    Bases(71) = 359
    Bases(72) = 367
    Bases(73) = 373
    Bases(74) = 379
    Bases(75) = 383
    Bases(76) = 389
    Bases(77) = 397
    Bases(78) = 401
    Bases(79) = 409
    Bases(80) = 419
    Bases(81) = 421
    Bases(82) = 431
    Bases(83) = 433
    Bases(84) = 439
    Bases(85) = 443
    Bases(86) = 449
    Bases(87) = 457
    Bases(88) = 461
    Bases(89) = 463
    Bases(90) = 467
    Bases(91) = 479
    Bases(92) = 487
    Bases(93) = 491
    Bases(94) = 499
    Bases(95) = 503
    Bases(96) = 509
    Bases(97) = 521
    Bases(98) = 523
    Bases(99) = 541
    PrimeInit = True
End Sub

'All the private functions that make the public functions work
Private Function prvAdd(a As HugeInt, b As HugeInt) As HugeInt
    'Debug.Print "prvAdd"
    Dim ua            As Long    'ubound a.digits
    Dim ub            As Long    'ubound b.digits
    Dim nd            As Long    'number of digits in shorter number
    Dim ne            As Long    'number of digits in longer number
    Dim i             As Long    'loop control variable
    Dim lPlaceHolder  As Long    'temp store for calcs
    Dim lPlaceHolder2 As Long    'second temp
    Dim Carry         As Boolean 'carry the one or not
    Dim iBuff()       As Integer 'buffer to store digits
    
    
    ua = UBound(a.Digits)
    ub = UBound(b.Digits)


    'Copy the overhang...
    '
    ' 123456789
    ' +   56789
    '-----------
    ' 1234
    
    If ua > ub Then
        ReDim iBuff(ua)
        nd = ub
        ne = ua
        For i = ua To ub + 1 Step -1
            iBuff(i) = a.Digits(i)
        Next
    Else
        ReDim iBuff(ub)
        nd = ua
        ne = ub
        For i = ub To ua + 1 Step -1
            iBuff(i) = b.Digits(i)
        Next
    End If




    'Use elementary school addition to add
    'other digits:
    '
    '    1      :leftover carry
    '     56789
    ' +   56789
    '-----------
    '     13578
    For i = 0 To nd
        lPlaceHolder = a.Digits(i) And &HFFFF& 'unsign a.Digits(i)
        lPlaceHolder2 = b.Digits(i) And &HFFFF& 'unsign b.Digits(i)
        lPlaceHolder = lPlaceHolder + lPlaceHolder2
        If Carry Then lPlaceHolder = lPlaceHolder + 1
        If lPlaceHolder And 65536 Then
            lPlaceHolder = lPlaceHolder - 65536
            Carry = True
        Else
            Carry = False
        End If
        
        'sign if necessary so it can return to an integer
        iBuff(i) = f(lPlaceHolder)
    Next




    'If carry1 is true then we have a leftover
    'carry from above:
    '
    '    1      :leftover carry
    ' 123413578 :current digits
    '-----------
    ' 123513578
    Do While Carry
        If i > ne Then
            'output isn't long enough
            ReDim Preserve iBuff(i)
        End If
        lPlaceHolder = iBuff(i) And &HFFFF& 'unsign
        lPlaceHolder = lPlaceHolder + 1
        If lPlaceHolder = 65536 Then
            Carry = True
            iBuff(i) = 0
        Else
            Carry = False
            iBuff(i) = f(lPlaceHolder)
        End If
        i = i + 1
    Loop

    prvAdd.Digits = TrimLeft(iBuff)
    
    Erase iBuff
End Function

Private Function prvDivide(Numerator As HugeInt, Denominator As HugeInt, _
    Remainder As HugeInt) As HugeInt

    Dim Quotient() As Integer
    Dim Shifted    As HugeInt
    Dim s          As HugeInt
    Dim i          As Long
    Dim j          As Long
    Dim n          As Long
    Dim u          As Long
    Dim t          As Long
    Dim p          As Long
    Dim x1         As Currency
    Dim x2         As Currency
    Dim x3         As Currency
    Dim Y          As Currency
    Dim y2         As Currency
    Dim y3         As Currency
    Dim c1         As Currency
    Dim c2         As Currency
    
    If IsOne(Denominator) Then
        Remainder = Zero
        prvDivide.Digits = Numerator.Digits
        Exit Function
    End If
    
    p = Compare(Numerator, Denominator)
    If p = Less Then
        Remainder.Digits = Numerator.Digits
        prvDivide = Zero
        Exit Function
    ElseIf p = Equal Then
        Remainder = Zero
        prvDivide = One
        Exit Function
    End If
    
    n = UBound(Numerator.Digits)
    t = UBound(Denominator.Digits)
    ReDim Quotient(n - t)
    Remainder.Digits = Numerator.Digits
    
'    Shifted = aShift(Denominator, n - t)
'
'    p = Compare(Remainder, Shifted)
'    y = (Denominator.Digits(t) And &HFFFF&) * CCur(65536)
'    If t > 0 Then y = y + (Denominator.Digits(t - 1) And &HFFFF&)
'    While p = Greater
'        x1 = (Remainder.Digits(n) And &HFFFF&) * CCur(65536)
'        If n > 0 Then x1 = x1 + (Remainder.Digits(n - 1) And &HFFFF&)
'        j = Int(x1 / y)
'        Quotient(n - t) = f((Quotient(n - t) And &HFFFF&) + j)
'        S = Zero
'        S.Digits(0) = j
'        S = Multiply(S, Shifted)
'        Remainder = prvSubtract(Remainder, S)
'        p = Compare(Remainder, Shifted)
'    Wend
    
    Y = Denominator.Digits(t) And &HFFFF&
    If t > 0 Then y2 = Denominator.Digits(t - 1) And &HFFFF& Else y2 = 0
    For i = n + 1 To t + 1 Step -1
        u = UBound(Remainder.Digits)
        If i > u Or i < 0 Then x1 = 0 Else x1 = Remainder.Digits(i) And &HFFFF&
        
        If i < u + 2 And i > 0 Then x2 = Remainder.Digits( _
            i - 1) And &HFFFF& Else x2 = 0
        
        If i < u + 3 And i > 1 Then x3 = Remainder.Digits( _
            i - 2) And &HFFFF& Else x3 = 0
        
        If i <= u Then
            If Remainder.Digits(i) = Denominator.Digits(t) Then
                p = 65535
            Else
                p = Int((x1 * 65536 + x2) / Y)
            End If
        Else
            p = Int((x1 * 65536 + x2) / Y)
        End If

        y3 = Y * 65536 + y2

        c1 = x1 * 4294967296# + x2 * 65536 + x3
        If y3 > 0 Then p = CLng(Int(c1 / y3))
        c2 = p * y3
        
        If c2 > c1 Then
            While c2 > c1
                p = p - 1
                c2 = p * y3
            Wend
        Else
            If y3 <= c1 - c2 Then
                p = p + 1
                c2 = p * y3
            End If
        End If
        
        s = IntOut(f(p))
        Shifted = aShift(Denominator, i - t - 1)
        s = Multiply(s, Shifted)
        
        Remainder = Subtract(Remainder, s)
        Quotient(i - t - 1) = f(p)
        
        While Remainder.IsNegative
            Remainder = Add(Remainder, Shifted)
            p = p - 1
            Quotient(i - t - 1) = f(p)
        Wend
        
    Next
    
    prvDivide.Digits = Quotient
    
    If IsEqual(Remainder, Denominator) Then
        prvDivide = Add(prvDivide, One)
        Remainder = Zero
    Else
        Remainder.Digits = TrimLeft(Remainder.Digits)
    End If

    Erase Quotient
    Erase s.Digits
    Erase Shifted.Digits
End Function

Private Function prvMontgomeryReduce(Number As HugeInt, Modulus As HugeInt) As HugeInt
    'Will only work when GCD(Number,modulus) = 1
    
    'Computes the Montgomery Reduction of Number mod Modulus
    'prvMontgomeryReduction = Number*Rinv mod Modulus
    '
    'Where: Rinv = modular inverse of B^(length of Modulus)
    
    'Dim R As HugeInt
    Dim mprime As HugeInt
    Dim a As HugeInt
    Dim u As Long
    Dim n As Long, i As Long
    
    n = UBound(Modulus.Digits) + 1
    If UBound(Number.Digits) > 2 * n - 1 Then Err.Raise 1001, "Montgomery Reduction", "Number can't be more than twice as long as the modulus."
    
    'R = aShift(One, n) ' R = base^n
    mprime = Equals(Modulus)
    mprime.IsNegative = True
    mprime = ModularInverse(mprime, aShift(One, 1))
    
    a = Equals(Number)
    ReDim Preserve a.Digits(2 * n - 1)
    
    For i = 0 To n - 1
        u = CLng(a.Digits(i)) * CLng(mprime.Digits(0) And &HFFFF&) Mod 65536
        a = Add(a, aShift(Multiply(IntOut(f(u)), Modulus), i))
    Next
    
    a = aShift(a, -n)
    If IsGreater(a, Modulus) Then a = Subtract(a, Modulus)
    
    prvMontgomeryReduce = Equals(a)
    
    Erase a.Digits
    Erase mprime.Digits

End Function

'Public Function InnerMultiply(a As HugeInt, b As HugeInt) As HugeInt
'    'Update: this function seems to be slower than traditional
'    '        multiplication, must be due to slow shifting and
'    '        partitioning
'
'
'    ' recursive multiplication function
'    ' O(n^(log 3 / log 2)), compared to O(n^2) of regular multiplication
'
'    ' This came to me when writing a routine for strassen's algorithm
'    ' for matrix multiplication. It's the same general idea
'
'    '   t1 = ab * ef
'    '   t2 = cd * gh
'    '   t3 = (ab + cd) * (ef + gh)
'    '   ...only 3 processor-intensive multiplications instead of 4
'    '   q=number of digits in abcd
'    '   abcd * efgh = t1*Base^q + (t3 - t1 - t2) * Base^(q/2) + t2
'    '   multiplying by the base is just a shift
'
'    ' We're going to pad the ends of the nmbers with 0's and this should
'    ' still be faster than traditional multiplication
'    Dim ua            As Long
'    Dim ub            As Long
'    Dim abcd          As HugeInt
'    Dim efgh          As HugeInt
'    Dim ab            As HugeInt
'    Dim cd            As HugeInt
'    Dim ef            As HugeInt
'    Dim gh            As HugeInt
'    Dim t1            As HugeInt
'    Dim t2            As HugeInt
'    Dim t3            As HugeInt
'    Dim cPlaceHolder  As Currency 'place holder is currency so we can multiply &H8000& * &H8000&
'    Dim lPlaceHolder  As Long
'    Dim lPlaceHolder2 As Long
'
'    ua = UBound(a.Digits)
'    ub = UBound(b.Digits)
'
'    If ua = 0 And ub = 0 Then
'        cPlaceHolder = CCur(a.Digits(0) And &HFFFF&) * CCur(b.Digits(0) And &HFFFF&)
'        lPlaceHolder = Int(cPlaceHolder / 65536)
'        lPlaceHolder2 = cPlaceHolder - CCur(65536) * CCur(lPlaceHolder)
'
'        If lPlaceHolder And &H8000 Then lPlaceHolder = lPlaceHolder - 65536
'        If lPlaceHolder2 And &H8000 Then lPlaceHolder2 = lPlaceHolder2 - 65536
'
'        If lPlaceHolder Then
'            ReDim ab.Digits(1)
'            ab.Digits(1) = lPlaceHolder
'        Else
'            ReDim ab.Digits(0)
'        End If
'        ab.Digits(0) = lPlaceHolder2
'
'        InnerMultiply.Digits = ab.Digits
'        Exit Function
'    End If
'
'
'    abcd.Digits = a.Digits
'    efgh.Digits = b.Digits
'
'    If ua > ub Then
'        If ua Mod 2 = 0 Then
'            ua = ua + 1
'            ReDim Preserve abcd.Digits(ua)
'        End If
'        ReDim Preserve efgh.Digits(ua)
'        ub = ua
'    ElseIf ub > ua Then
'        If ub Mod 2 = 0 Then
'            ub = ub + 1
'            ReDim Preserve efgh.Digits(ub)
'        End If
'        ReDim Preserve abcd.Digits(ub)
'        ua = ub
'    ElseIf ua Mod 2 = 0 Then
'        ua = ua + 1
'        ReDim Preserve abcd.Digits(ua)
'        ReDim Preserve efgh.Digits(ua)
'        ub = ua
'    End If
'    ub = ub + 1
'    ab.Digits = ArrayPartition(abcd.Digits, ub / 2, ua)
'    ef.Digits = ArrayPartition(efgh.Digits, ub / 2, ua)
'
'    cd.Digits = ArrayPartition(abcd.Digits, 0, ub / 2 - 1)
'    gh.Digits = ArrayPartition(efgh.Digits, 0, ub / 2 - 1)
'
'    t1.Digits = InnerMultiply(ab, ef).Digits
'    t2.Digits = InnerMultiply(cd, gh).Digits
'    t3.Digits = InnerMultiply(prvAdd(ab, cd), prvAdd(ef, gh)).Digits
'    t3.Digits = prvSubtract(prvSubtract(t3, t1), t2).Digits
'
'    ' adding t1*Base^q + (t3-t1-t2)*Base^(q/2) + t2 without having to shift
'    ' bits all over the place is possible, but I'm too lazy right now...
'
'    Dim i As Long
'    Dim ua2 As Long
'    ua2 = (ua - 1) / 2
'    ReDim Preserve t1.Digits(UBound(t1.Digits) + ub)
'    ReDim Preserve t3.Digits(UBound(t3.Digits) + ua2 + 1)
'
'    For i = UBound(t3.Digits) - 1 To ua2 Step -1
'        t3.Digits(i + 1) = t3.Digits(i - ua2)
'    Next
'    For i = UBound(t1.Digits) - 1 To ua Step -1
'        t1.Digits(i + 1) = t1.Digits(i - ua)
'    Next
'    For i = 0 To ua2
'        t3.Digits(i) = 0
'        t1.Digits(2 * i) = 0
'        t1.Digits(2 * i + 1) = 0
'    Next
''    For i = UBound(t1.Digits) To 0 Step -1
''        'Debug.Print Format(t1.Digits(i), "000") & " ";
''    Next
''    'Debug.Print
''    For i = UBound(t3.Digits) To 0 Step -1
''        'Debug.Print Format(t3.Digits(i), "000") & " ";
''    Next
''    'Debug.Print
''    For i = UBound(t2.Digits) To 0 Step -1
''        'Debug.Print Format(t2.Digits(i), "000") & " ";
''    Next
''    'Debug.Print
'    InnerMultiply.Digits = prvAdd(prvAdd(t1, t3), t2).Digits
'
'    InnerMultiply.Digits = TrimLeft(InnerMultiply.Digits)
'
'    Erase t1.Digits
'    Erase t2.Digits
'    Erase t3.Digits
'    Erase ab.Digits
'    Erase cd.Digits
'    Erase ef.Digits
'    Erase gh.Digits
'    Erase abcd.Digits
'    Erase efgh.Digits
'
'End Function

Private Function prvMultiply(a As HugeInt, b As HugeInt) As HugeInt
    'Debug.Print "prvMultiply"

    Dim i As Long, j As Long, k As Long
    Dim cPlaceHolder As Currency
    Dim lPlaceHolder As Long
    Dim lPlaceHolder2 As Long
    Dim Carry As Long
    Dim iBuff() As Integer
    
    ReDim iBuff(UBound(a.Digits) + UBound(b.Digits) + 1)
    
    For i = 0 To UBound(a.Digits)
        lPlaceHolder = a.Digits(i) And &HFFFF&
        If lPlaceHolder <> 0 Then
            For j = 0 To UBound(b.Digits)
                lPlaceHolder2 = b.Digits(j) And &HFFFF&
                cPlaceHolder = CCur(lPlaceHolder) * CCur(lPlaceHolder2)
    
                lPlaceHolder2 = iBuff(i + j) And &HFFFF&
                cPlaceHolder = lPlaceHolder2 + cPlaceHolder
                Carry = Int(cPlaceHolder / 65536)
                cPlaceHolder = cPlaceHolder - CCur(Carry) * 65536
                
                If cPlaceHolder > 32767 Then
                    iBuff(i + j) = cPlaceHolder - 65536
                Else
                    iBuff(i + j) = cPlaceHolder
                End If
                
                k = 0
                Do
                    k = k + 1
                    lPlaceHolder2 = iBuff(i + j + k) And &HFFFF&
                    lPlaceHolder2 = lPlaceHolder2 + Carry
                    If lPlaceHolder2 > 65535 Then
                        lPlaceHolder2 = lPlaceHolder2 - 65536
                        Carry = 1
                        iBuff(i + j + k) = f(lPlaceHolder2)
                    Else
                        iBuff(i + j + k) = f(lPlaceHolder2)
                        Exit Do
                    End If
                Loop
            Next
        End If
    Next
    If iBuff(UBound(iBuff)) = 0 Then ReDim Preserve iBuff(UBound(iBuff) - 1)
    prvMultiply.Digits = iBuff
    Erase iBuff
End Function

Private Function prvPower(Base As HugeInt, Power As HugeInt) As HugeInt
    'Debug.Print "prvPower"

    Dim tmpSquare As HugeInt
    Dim Output As HugeInt
    Dim j As Long
    Dim i As Long
    Dim k As Single
    
    tmpSquare.Digits = Base.Digits
    ReDim Output.Digits(0)
    Output.Digits(0) = 1
    
    For i = 0 To UBound(Power.Digits) - 1
        For j = 0 To 15
            If (i Or j) Then tmpSquare = prvSquare(tmpSquare)
            If Power.Digits(i) And 2 ^ j Then
                Output = Multiply(Output, tmpSquare)
            End If
        Next
    Next
    
    If Power.Digits(i) And &HFFFF& > 0 Then
        k = Int(Log(Power.Digits(i) And &HFFFF&) / Log(2))
    
        For j = 0 To k
            If (i Or j) Then tmpSquare = prvSquare(tmpSquare)
            If Power.Digits(i) And 2 ^ j Then
                Output = Multiply(Output, tmpSquare)
            End If
        Next
    End If
    prvPower.Digits = Output.Digits
    
    Erase Output.Digits
    Erase tmpSquare.Digits
End Function

Private Function prvSquare(a As HugeInt) As HugeInt
    'Debug.Print "prvSquare"

    'Squares argument a whole lot faster than Multiply(a, a)
    Dim iBuff() As Long
    Dim iBuff2() As Integer
    Dim ua As Long
    Dim cPlaceHolder As Currency
    Dim lPlaceHolder As Long
    Dim lPlaceHolder2 As Long
    Dim Carry As Long
    Dim i As Long
    Dim j As Long
    
    ua = UBound(a.Digits) + 1
    
    ReDim iBuff(2 * ua - 1)
    
    For i = 0 To ua - 1
        lPlaceHolder = iBuff(2 * i)
        lPlaceHolder2 = a.Digits(i) And &HFFFF&
        cPlaceHolder = CCur(lPlaceHolder) + CCur(lPlaceHolder2) * CCur( _
            lPlaceHolder2)
        Carry = Int(cPlaceHolder / 65536)
        lPlaceHolder = cPlaceHolder - CCur(Carry) * 65536
        iBuff(2 * i) = lPlaceHolder
        For j = i + 1 To ua - 1
            cPlaceHolder = CCur(iBuff(i + j)) + 2 * CCur(a.Digits( _
                j) And &HFFFF&) * CCur(a.Digits(i) And &HFFFF&) + Carry
            Carry = Int(cPlaceHolder / 65536)
            lPlaceHolder = cPlaceHolder - CCur(Carry) * 65536
            iBuff(i + j) = lPlaceHolder
        Next
        'lPlaceHolder = Carry Mod 65536
        iBuff(i + ua) = Carry
        'lPlaceHolder = Carry \ 65536
        'If lPlaceHolder Then iBuff(i + ua + 1) = f(lPlaceHolder)
    Next
    
    If iBuff(2 * ua - 1) = 0 Then ReDim Preserve iBuff(2 * ua - 2)
    
    ReDim iBuff2(UBound(iBuff))
    For i = 0 To UBound(iBuff)
        iBuff2(i) = f(iBuff(i))
    Next
    
    prvSquare.Digits = iBuff2
    Erase iBuff
    Erase iBuff2

End Function

Private Function prvSubtract(Greater As HugeInt, Smaller As HugeInt) As HugeInt
    'Debug.Print "prvSubtract"
    Dim ug            As Long
    Dim us            As Long
    Dim nd            As Long
    Dim ne            As Long
    Dim i             As Long
    Dim lPlaceHolder  As Long
    Dim lPlaceHolder2 As Long
    Dim Carry         As Boolean
    Dim iBuff()       As Integer
    
    ug = UBound(Greater.Digits)
    us = UBound(Smaller.Digits)
    
    'Copy the overhang...
    '
    ' 123456789
    ' +   56789
    '-----------
    ' 1234
    nd = us
    If ug > us Then
        ne = ug
        ReDim iBuff(ne)
        For i = ug To us + 1 Step -1
            iBuff(i) = Greater.Digits(i)
        Next
    Else
        ne = us
        ReDim iBuff(ne)
    End If



    'Subtract the digits that line up
    For i = 0 To nd
        lPlaceHolder = Greater.Digits(i) And &HFFFF&
        lPlaceHolder2 = Smaller.Digits(i) And &HFFFF&
        lPlaceHolder = lPlaceHolder - lPlaceHolder2
        If Carry Then lPlaceHolder = lPlaceHolder - 1
        If lPlaceHolder < 0 Then
            lPlaceHolder = lPlaceHolder + 65536
            Carry = True
        Else
            Carry = False
        End If
        iBuff(i) = f(lPlaceHolder)
    Next


    'if carry is true subtract one from each digit
    'until we get to one that isn't 0
    
    If Carry Then
        Do
            If iBuff(i) = 0 Then
                iBuff(i) = -1
            Else
                lPlaceHolder = iBuff(i) And &HFFFF&
                lPlaceHolder = lPlaceHolder - 1
                iBuff(i) = f(lPlaceHolder)
                Exit Do
            End If
            i = i + 1
        Loop
    End If
    
    If iBuff(ne) = 0 And ne > 0 Then ReDim Preserve iBuff(ne - 1)
    
    prvSubtract.Digits = TrimLeft(iBuff)
    
    Erase iBuff
End Function

Private Function TrimLeft(a() As Integer) As Integer()
    Dim i As Long
    i = UBound(a)
    Do Until Not (a(i) = 0)
        i = i - 1
        If i = -1 Then
            i = 0
            Exit Do
        End If
    Loop
    TrimLeft = ArrayPartition(a, 0, i)
End Function
#End If
