Attribute VB_Name = "Primes"
#If APM Then
Option Explicit
Private Declare Function mod64 Lib "ntdll.dll" Alias "_aullrem" ( _
ByVal c64Dividend As Currency, _
ByVal iDivisorLo As Long, _
ByVal iDivisorHi As Long) As Long

Private Declare Function allmul Lib "msvbvm60.dll" Alias "_allmul" (ByVal curA As Currency, _
ByVal curB As Currency) As Currency
'Primality testing

'//////////////////////////////////////////////////////////////////////////////////////////////
'/// Copyright (c) Andrew Mallner, All Rights Reserved                                      ///
'/// [partially modified by OnError]                                                        ///
'/// this code may be used for educational purposes only, and may not be redistributed      ///
'///                                                                                        ///
'/// THIS CODE IS ONLY USED TO SUPPORT THE "Factor" FUNCTION IN modArbitraryPrecision       ///
'/// IT IS NOT NECESSARY FOR ANY OTHER PART OF THAT MODULE                                  ///
'//////////////////////////////////////////////////////////////////////////////////////////////

Public Function sIsPrime(ByVal Number As Long) As Boolean
    '/////////////////////////////////////////////////////////
    '///  clsMath.IsPrime                                  ///
    '///                                                   ///
    '///  Purpose - Uses Fermat's Little Theorem to        ///
    '///            determine whether the argument is      ///
    '///            prime                                  ///
    '/////////////////////////////////////////////////////////

    If (Number And 1) = 0 And Number > 2 Then Exit Function

    'If number = 2 Then IsPrime = True: Exit Function
    'I don't know why OnErr0r added the above line
    'it may speed up isprime(2), but it slows down
    'isprime for ALL other arguments, considering
    'the same comparison is made inside the loop

    Dim i As Long, n As Long, o As Long, x(4) As Long
    x(0) = 2: x(1) = 3: x(2) = 5: x(3) = 7: x(4) = 11

    If Number > 46340 Then
        For i = 0 To 4
            o = x(i)
            If Number = o Then i = 5: Exit For
            n = sModularExponentiationOE(o, CLng(Number) - 1, Number)
            If n <> 1 And n <> Number - 1 Then Exit For
        Next
    Else
        For i = 0 To 4
            o = x(i)
            If Number = o Then i = 5: Exit For
            n = sModularExponentiation(o, CLng(Number) - 1, Number)
            If n <> 1 And n <> Number - 1 Then Exit For
        Next
    End If
    
    If i = 5 Then sIsPrime = True
End Function

Public Function sNextPrime(ByVal Number As Double) As Double
    '/////////////////////////////////////////////////////////
    '///                                                   ///
    '///  NextPrime                                        ///
    '///                                                   ///
    '///  Purpose - Return the least prime greater than    ///
    '///            the argument                           ///
    '///                                                   ///
    '/////////////////////////////////////////////////////////

    Number = Number + 1
    Do Until sIsPrime(Number)
        Number = Number + 1
    Loop
    sNextPrime = Number
End Function

Public Function sPrevPrime(ByVal Number As Double) As Double
    '/////////////////////////////////////////////////////////
    '///                                                   ///
    '///  PrevPrime                                        ///
    '///                                                   ///
    '///  Purpose - Return the greatest prime less than    ///
    '///            the argument                           ///
    '///                                                   ///
    '/////////////////////////////////////////////////////////

    Number = Number - 1
    Do Until sIsPrime(Number)
        Number = Number - 1
        If Number < 2 Then sPrevPrime = 0: Exit Function
    Loop
    sPrevPrime = Number
End Function

'Public Function ModularExponentiation(ByVal Base As Long, ByVal Exponent As Long, ByVal Modulus As Long) As Double
'    '/////////////////////////////////////////////////////////
'    '///  clsMod.ModularExponentiation                     ///
'    '///                                                   ///
'    '///  Purpose - calculate (base^exponent) mod modulus  ///
'    '///                                                   ///
'    '///  Overflow problem fixed, thanks to MathImagics    ///
'    '///                                                   ///
'    '/////////////////////////////////////////////////////////
'
'    Dim i As Long
'    Dim b(31) As Byte
'    Dim t As Double
'    Dim lMaxBit As Long
'
'    lMaxBit = DecToBin(Exponent, b)
'    t = Base Mod Modulus
'
'    If b(0) Then
'        ModularExponentiation = t
'    Else
'        ModularExponentiation = 1
'    End If
'    If Exponent <> 0 Then
'      For i = 1 To lMaxBit
'          t = MulMod32(t, t, Modulus)
'          If b(i) Then
'            ModularExponentiation = MulMod32(t, ModularExponentiation, Modulus)
'          End If
'      Next
'   End If
'End Function

Private Function DecToBin(ByVal lDecimal As Long, ByRef b() As Byte)
    '/////////////////////////////////////////////////////////
    '///                                                   ///
    '///  DecToBin                                         ///
    '///                                                   ///
    '///  Purpose - returns a byte array representing      ///
    '///            ldecimal in binary                     ///
    '///                                                   ///
    '///  Written by OnErr0r                               ///
    '///                                                   ///
    '/////////////////////////////////////////////////////////
  Dim lPos As Long
  
  Do
    b(lPos) = (lDecimal And 1)
    lDecimal = lDecimal \ 2
    lPos = lPos + 1
  Loop While lDecimal
  
  DecToBin = lPos - 1
End Function

Public Function MulMod32(ByVal a As Double, ByVal b As Double, ByVal M As Double)
    '
    '  Returns A * B mod M for any positive A, B, M
    '      MathImagics RSA 32-bit Library
    '      by Dr Memory
    '
    '  Guaranteed not to overflow!
    '
    '  optimized slightly by Andrew Mallner
    '

    Const MAXLONG32 = &H7FFFFFFF

    Dim MM As Long

    a = a Mod M
    If Abs(a * b) > MAXLONG32 Then
        While b > 0
            If (b Mod 2) = 1 Then
                If a > MAXLONG32 - MM Then   ' (A + MM) Mod M  will overflow
                    If a <= MM Then MM = a - (M - MM) Else MM = MM - (M - a)
                Else
                    MM = (a + MM) Mod M     ' it's safe
                End If
            End If
                    
            If a > MAXLONG32 - a Then        ' ditto for 2*A mod M
                a = a - (M - a)
            Else
                a = (a + a) Mod M
            End If
            b = b \ 2
        Wend
    Else
        MM = a * b
        MM = MM - Int(MM / M) * M
    End If
    MulMod32 = MM
End Function

Public Function sFactor(ByVal Number As Double) As Double()
    '/////////////////////////////////////////////////////////
    '///                                                   ///
    '///  Factor                                           ///
    '///                                                   ///
    '///  Purpose - Uses Pollard's Rho method to factor    ///
    '///            numbers into an array of primes        ///
    '///                                                   ///
    '/////////////////////////////////////////////////////////


    'Don't try to facor a prime
    If sIsPrime(Number) Then
        Dim d(0) As Double
        d(0) = Number
        sFactor = d
        Exit Function
    End If
    
'    'Function sucks at factoring 4
'    If Number = 4 Then
'        Dim c(1) As Double
'        c(0) = 2
'        c(1) = 2
'        Factor = c
'        Exit Function
'    End If
'
'    'Function sucks at factoring 9
'    If Number = 9 Then
'        Dim b(1) As Double
'        b(0) = 3
'        b(1) = 3
'        Factor = b
'        Exit Function
'    End If
    
    Randomize
    Dim x(254) As Long, seed As Long, i As Byte, g As Double, j As Integer, k As Integer, l As Integer
    Dim f() As Double, e() As Double, t As Double, q As Integer, r As Integer, s As Integer
    j = 1
    l = -1
    x(0) = Int(Rnd * 1000) + 1
    q = Int(Rnd * 2) + 1
    r = Int(Rnd * 2) + 1
    s = Int(Rnd * 2) + 1
    For i = 1 To 254
        'need slight randomization in generating polynomial
        'if x^2 - x + 1 fails once, it will continue to fail
        'until we run out of stack space.
        
        't = (q * x(i - 1) ^ 2 - r * x(i - 1) + s)
        If x(i - 1) > 46340 Or x(i - 1) < -46340 Then
            t = RhoPoly(q, r, s, x(i - 1), CLng(Number))
            't = MulMod32(MulMod32(x(i - 1), x(i - 1), Number), t, Number)
            't = (t - r * x(i - 1) + s) Mod Number
        Else
            t = (x(i - 1) * x(i - 1)) Mod Number
            t = (t * q - r * x(i - 1) + s) Mod Number
        End If
        
        
        x(i) = t ' - Int(t / number) * number
        If i = 2 * j Then
            If x(i) = x(j) Then Exit For
            g = sGCD(Abs(x(i) - x(j)), Number)
            
            If g > 1 And g < Number Then
                Number = Number / g
                e = sFactor(g)
                ReDim Preserve f(l + UBound(e) + 1)
                For k = 0 To UBound(e)
                    l = l + 1
                    f(l) = e(k)
                Next
                Exit For
            End If
            
            j = j + 1
        End If
    Next
    e = sFactor(Number)
    ReDim Preserve f(l + UBound(e) + 1)
    For k = 0 To UBound(e)
        l = l + 1
        f(l) = e(k)
    Next
    sFactor = f
End Function

Public Function sGCD(ByVal a As Double, ByVal b As Double) As Double
    '/////////////////////////////////////////////////////////
    '///                                                   ///
    '///  clsMath.GCD                                      ///
    '///                                                   ///
    '///  Purpose - Return the greatest common divisor     ///
    '///            of a and b                             ///
    '///                                                   ///
    '/////////////////////////////////////////////////////////

    Do Until a = 0 Or b = 0
    If a > b Then
        'a = a Mod b
        a = a - Int(a / b) * b
    Else
        'b = b Mod a
        b = b - Int(b / a) * a
    End If
    Loop
    
    If a = 0 Then sGCD = b Else sGCD = a
End Function

Public Function sModularExponentiation(ByVal Base As Long, ByVal Exponent As Long, ByVal Modulus As Long) As Long
    '/////////////////////////////////////////////////////////
    '///  clsMod.ModularExponentiation                     ///
    '///                                                   ///
    '///  Purpose - calculate (base^exponent) mod modulus  ///
    '///  Warning - overflow can occur if modulus is       ///
    '///            greater than 46341                     ///
    '/////////////////////////////////////////////////////////
    
    Dim i As Long
    Dim b(31) As Byte
    Dim t As Currency
    Dim lMaxBit As Long

    lMaxBit = DecToBin(Exponent, b)

    t = Base Mod Modulus
    If b(0) Then
        sModularExponentiation = t
    Else
        sModularExponentiation = 1
    End If
    
    If Exponent <> 0 Then
      For i = 1 To lMaxBit
          t = t * t
          t = t - Int(t / Modulus) * Modulus
          If b(i) Then
            sModularExponentiation = t * sModularExponentiation
            sModularExponentiation = sModularExponentiation Mod Modulus
            'ModularExponentiation = ModularExponentiation - Int(ModularExponentiation / Modulus) * Modulus
          End If
      Next
   End If
End Function

Public Function sModularExponentiationOE(ByVal Base As Long, ByVal Exponent As Long, ByVal Modulus As Long) As Currency
    '/////////////////////////////////////////////////////////
    '///  clsMod.ModularExponentiation                     ///
    '///                                                   ///
    '///  Purpose - calculate (base^exponent) mod modulus  ///
    '///  Warning - overflow can occur if modulus is       ///
    '///            greater than 46341                     ///
    '/////////////////////////////////////////////////////////
    
    Dim i As Long
    Dim b(31) As Byte
    Dim t As Currency
    Dim lMaxBit As Long

    lMaxBit = DecToBin(Exponent, b)

    t = Base Mod Modulus
    If b(0) Then
        sModularExponentiationOE = t
    Else
        sModularExponentiationOE = 1
    End If
    
    If Exponent <> 0 Then
      For i = 1 To lMaxBit
          t = allmul(t / 10000, t / 10000)
          't = t * t
          t = mod64(t, Modulus, 0)
          't = t - Int(t / Modulus) * Modulus
          If b(i) Then
            'ModularExponentiationOE = t * ModularExponentiationOE
            sModularExponentiationOE = allmul(t / 10000, sModularExponentiationOE / 10000)
            sModularExponentiationOE = mod64(sModularExponentiationOE, Modulus, 0)
            'ModularExponentiationOE = ModularExponentiationOE - Int(ModularExponentiationOE / Modulus) * Modulus
          End If
      Next
   End If
End Function

Public Function MulMod(a As Long, b As Long, Modulus As Long) As Currency

MulMod = allmul(CCur(a) / 10000, CCur(b) / 10000)
MulMod = mod64(MulMod, Modulus, 0)

End Function

Public Function RhoPoly(q As Integer, r As Integer, s As Integer, x As Long, n As Long) As Currency

RhoPoly = mod64(allmul(allmul(CCur(x) / 10000, CCur(x) / 10000) / 10000, CCur(q) / 10000) - allmul(CCur(x), CCur(r)) + s, n, 0)

End Function
#End If
