Attribute VB_Name = "modStrMath"
Option Explicit

Public Function StrAdd(sA As String, sB As String) As String
    Dim i As Integer
    Dim iLenA As Integer, iLenB As Integer
    Dim iTemp As Integer, iCarry As Integer
    Dim sOut As String
        
    'Calculate once and store, more efficient
    iLenA = Len(sA)
    iLenB = Len(sB)
    
    'Pad out the shorter string with leading 0s
    If (iLenA > iLenB) Then
        sB = String$(iLenA - iLenB, "0") & sB
    ElseIf (iLenB > iLenA) Then
        sA = String$(iLenB - iLenA, "0") & sA
        iLenA = iLenB
    End If
    
    'Allocate string space for the result
    sOut = Space$(iLenA)
    
    For i = iLenA To 1 Step -1
        'Add them together. 48 is the ASCII code for 0, so we subtract
        '96 which is 2*48 to basically convert the returned Asc values
        'into the actual numbers they represent
        iTemp = Asc(Mid$(sA, i, 1)) + Asc(Mid$(sB, i, 1)) + iCarry - 96
        
        'We use Mod 10 so that it only returns the last digit if the number is greater than 9
        Mid$(sOut, i, 1) = iTemp Mod 10
        
        'We use integer divide so that it always rounds down, so 9 \ 10 = 0 instead of 1
        iCarry = iTemp \ 10
    Next i
    
    If iCarry Then sOut = "1" & sOut

    'Return the result
    StrAdd = sOut
End Function

Public Function StrSub(sA As String, sB As String) As String
    Dim i As Integer, j As Integer
    Dim iLenA As Integer, iLenB As Integer
    Dim iTemp As Integer, iTemp2 As Integer
    Dim sOut As String
        
    'Calculate once and store, more efficient
    iLenA = Len(sA)
    iLenB = Len(sB)
    
    'Pad out the shorter string with leading 0s
    If (iLenA > iLenB) Then
        sB = String$(iLenA - iLenB, "0") & sB
    ElseIf (iLenB > iLenA) Then
        sA = String$(iLenB - iLenA, "0") & sA
        iLenA = iLenB
    End If
    
    'Allocate string space for the result
    sOut = Space$(iLenA)
    
    
    For i = iLenA To 1 Step -1
        'Basic subtraction. We dont need to worry about subtracting 48 or 96
        'since they cancel each other out. 5 - 3 is the same as 53 - 51 (the ASCII codes)
        iTemp = Asc(Mid$(sA, i, 1)) - Asc(Mid$(sB, i, 1))
        
        'If the number is less than 0, we need to break some greater number down
        'into units. Basic school maths here.
        If (iTemp < 0) Then
        
            'Find a number which is not 0
            For j = i - 1 To 1 Step -1
                iTemp2 = Asc(Mid$(sA, j, 1))
                If iTemp2 > 48 Then
                    Mid$(sA, j, 1) = Chr$(iTemp2 - 1)
                    Exit For
                End If
            Next j
            
            If j = 0 Then
                'Out of numbers, and Im not going to deal with negatives
                'at all, so just raise an error and exit
                Err.Raise vbObjectError + 100, , "String math does not support negative numbers"
                Exit Function
            End If
            
            'If you understand the maths you'll know why I do this.
            'If you dont, read a book, go to school, or whatever
            For j = j + 1 To i
                Mid$(sA, j, 1) = "9"
            Next j
            
            iTemp = iTemp + 10
        End If

        'Fill the result string
        Mid$(sOut, i, 1) = iTemp
    Next i
    
    'Return the result
    StrSub = sOut
End Function

Public Function StrMod(sA As String, sB As String) As String
    ' sB will be trimed as a side effect of this function
    ' sA better not contain any leading zeros to get faster results
    
    Dim sCurrent As String
    Dim sB2 As String
    Dim lTemp As Long
    
    sCurrent = sA: sB = StrTrim(sB): lTemp = Len(sCurrent) - Len(sB)
    ' Multiply sB by 10^ltemp.
    sB2 = sB & String(IIf(lTemp > 0, lTemp, 0) + 1, 48)
    Do
        ' Devide over 10
        sB2 = Left(sB2, Len(sB2) - 1)
        
        While StrGT(sCurrent, sB2)
            sCurrent = StrSub(sCurrent, sB2)
        Wend
        sB2 = StrTrim(sB2)
    Loop While sB2 <> sB
    
    sCurrent = StrTrim(sCurrent)
    StrMod = IIf(sCurrent = sB, 0, sCurrent)
End Function

Public Function StrDiv2(sA As String, sB As String) As String
    ' sB will be trimed as a side effect of this function
    ' sA better not contain any leading zeros to get faster results
    
    Dim sCurrent As String, sCount As String
    Dim sB2 As String
    Dim lTemp As Long, lCount As Long
    
    sCount = "0"
    sCurrent = sA: sB = StrTrim(sB): lTemp = Len(sCurrent) - Len(sB)
    lTemp = IIf(lTemp > 0, lTemp, 0) + 1
    ' Multiply sB by 10^ltemp.
    sB2 = sB & String$(lTemp, 48)
    Do
        ' Devide over 10
        sB2 = Left(sB2, Len(sB2) - 1): lTemp = lTemp - 1
        
        While Not StrGT(sB2, sCurrent)
            sCurrent = StrSub(sCurrent, sB2)
            sCount = StrAdd(sCount, "1" & String$(lTemp, 48))
        Wend
        sB2 = StrTrim(sB2)
    Loop While sB2 <> sB

    StrDiv2 = sCount
End Function
Public Function StrMult(sA As String, sB As String) As String
    Dim i As Integer, j As Integer
    Dim iLenA As Integer, iLenB As Integer
    Dim iTemp As Integer, iCarry As Integer
    Dim sOut As String, sTemp As String
    
    'Again, I'm using the math tecniques I got taught at school
    'Namely, long multiplication
    
    'Calculate once and store, more efficient
    iLenA = Len(sA)
    iLenB = Len(sB)
    
    'Pad out the shorter string with leading 0s
    If (iLenA > iLenB) Then
        sB = String$(iLenA - iLenB, "0") & sB
    ElseIf (iLenB > iLenA) Then
        sA = String$(iLenB - iLenA, "0") & sA
        iLenA = iLenB
    End If
    
    'We dont preallocate space because we'll be calling the StrAdd function later
    
    For i = iLenA To 1 Step -1
        'This was missing.
        'The carry-variable should be emptied when going to a new 'line'
        iCarry = 0
    
        'Preallocate space for this 'line'
        sTemp = Space(iLenA + 1)
        For j = iLenA To 1 Step -1
            
            'Calculate
            iTemp = ((Asc(Mid$(sA, i, 1)) - 48) * (Asc(Mid$(sB, j, 1)) - 48)) + iCarry
            'Save the 'units' portion
            Mid$(sTemp, j, 1) = iTemp Mod 10
            'Carry the 'tens' portion
            iCarry = iTemp \ 10
        Next j
        
        If iCarry Then sTemp = iCarry & sTemp
        'Multiply by 10^digitnumber (from the right) then add
        sOut = StrAdd(sOut, Trim$(sTemp) & String$(iLenA - i, "0"))
        DoEvents
    Next i
    
    StrMult = StrTrim(sOut)
End Function
Public Function StrDiv(sA As String, sB As String) As String
    'Again, a fairly sloppy solution, I am going to subtract sB from sA
    'until sA < sB and count how many times I've done it
    Dim sCurrent As String
    Dim sCount As String
    Dim lCount As Long
    
    sCurrent = sA
    While Not (StrGT(sB, sCurrent))
        sCurrent = StrSub(sCurrent, sB)
        lCount = lCount + 1
        If lCount > 2147483640 Then
            sCount = StrAdd(sCount, CStr(lCount))
            lCount = 0
        End If
    Wend
        
    If (lCount) Then sCount = StrAdd(sCount, CStr(lCount))
    StrDiv = sCount
End Function

Public Function StrPow(sA As String, sB As String) As String
    Dim sOut As String
    Dim sCurrent As String
    Dim sTemp As String
    
    sOut = sA
    sCurrent = "1"
    
    'A series of multiplications, either squares (current*current)
    'or doubles (current*x)
    
    Do
        sTemp = StrMult(sCurrent, "2")
        If StrGT(sTemp, sB) Then
            sOut = StrMult(sOut, sA)
            sCurrent = StrAdd(sCurrent, "1")
        Else
            sOut = StrMult(sOut, sOut)
            sCurrent = sTemp
        End If
        DoEvents 'Need this, the process can be quite long!
        Debug.Print sCurrent
    Loop Until sCurrent = sB
    
    StrPow = sOut
End Function

Public Function StrGT(sA As String, sB As String) As Boolean
    Dim i As Integer
    Dim iLenA As Integer, iLenB As Integer
    Dim iTemp As Integer, iTemp2 As Integer
        
    'Calculate once and store, more efficient
    iLenA = Len(sA)
    iLenB = Len(sB)
    
    'Pad out the shorter string with leading 0s
    If (iLenA > iLenB) Then
        sB = String$(iLenA - iLenB, "0") & sB
    ElseIf (iLenB > iLenA) Then
        sA = String$(iLenB - iLenA, "0") & sA
        iLenA = iLenB
    End If
    
    'You're thinking : surely if sA is a longer string than sB then it must be greater?
    'The reason is I want to allow for any number of leading 0s
    
    'Check. As soon as the numbers dont match we can say which one is greater
    For i = 1 To iLenA
        iTemp = Asc(Mid$(sA, i, 1))
        iTemp2 = Asc(Mid$(sB, i, 1))
        
        If iTemp > iTemp2 Then
            'sA is greater
            StrGT = True
            Exit Function
        ElseIf iTemp2 > iTemp Then
            'sB is greater
            Exit For
        End If
    Next i
    
    'Return false
    StrGT = False
End Function

Public Function StrTrim(sA As String) As String
    ' Zero or Empty Strings will return 0
    
    Dim i As Integer
    
    i = 1
    On Error GoTo er1
    Do Until Asc(Mid$(sA, i, 1)) > 48
        i = i + 1
    Loop
    On Error GoTo 0
    StrTrim = Mid$(sA, i): Exit Function

er1:
    StrTrim = "0": Resume out
out:
End Function

